"""Testing xonsh json hooks"""
import json

import pytest

from xonsh.jsonutils import serialize_xonsh_json
from xonsh.tools import EnvPath


@pytest.mark.parametrize(
    "inp",
    [
        42,
        "yo",
        ["hello"],
        {"x": 65},
        EnvPath(["wakka", "jawaka"]),
        ["y", EnvPath(["wakka", "jawaka"])],
        {"z": EnvPath(["wakka", "jawaka"])},
    ],
)
def test_serialize_xonsh_json_roundtrip(inp):
    s = json.dumps(inp, default=serialize_xonsh_json)
    obs = json.loads(s)
    assert inp == obs



from xonsh.cli_utils import ArgparseCompleter
from xonsh.parsers.completion_context import CommandContext


def xonsh_complete(command: CommandContext):
    """Completer for ``xonsh`` command using its ``argparser``"""

    from xonsh.main import parser

    completer = ArgparseCompleter(parser, command=command)
    return completer.complete(), False



"""Test XonshLexer for pygments"""


import pytest
from pygments.token import (
    Error,
    Keyword,
    Literal,
    Name,
    Number,
    Operator,
    Punctuation,
    String,
    Text,
)

from xonsh.environ import LsColors
from xonsh.events import EventManager, events
from xonsh.pyghooks import Color, XonshLexer, XonshStyle, on_lscolors_change
from xonsh.pytest.tools import DummyShell, skip_if_on_windows


@pytest.fixture
def xsh(xession, monkeypatch):
    for key in ("cd", "bash"):
        monkeypatch.setitem(xession.aliases, key, lambda *args, **kwargs: None)


@pytest.fixture()
def check_token(xsh):
    def factory(code, tokens):
        """Make sure that all tokens appears in code in order"""
        lx = XonshLexer()
        tks = list(lx.get_tokens(code))

        for tk in tokens:
            while tks:
                if tk == tks[0]:
                    break
                tks = tks[1:]
            else:
                msg = f"Token {tk!r} missing: {list(lx.get_tokens(code))!r}"
                pytest.fail(msg)
                break

    return factory


_cases = {
    "ls": {
        "ls -al": [
            (Name.Builtin, "ls"),
        ],
    },
    "ls-bin": {
        "/bin/ls -al": [
            (Name.Builtin, "/bin/ls"),
        ],
    },
    "print": {
        'print("hello")': [
            (Name.Builtin, "print"),
            (Punctuation, "("),
            (Literal.String.Double, '"'),
            (Literal.String.Double, "hello"),
            (Literal.String.Double, '"'),
            (Punctuation, ")"),
            (Text, "\n"),
        ]
    },
    "invalid-cmd": {
        "non-existance-cmd -al": [
            (Name, "non"),
        ],
        "![non-existance-cmd -al]": [
            (Error, "non-existance-cmd"),
        ],
        "for i in range(10):": [
            (Keyword, "for"),
        ],
        "(1, )": [
            (Punctuation, "("),
            (Number.Integer, "1"),
        ],
    },
    "multi-cmd": {
        "cd && cd": [
            (Name.Builtin, "cd"),
            (Operator, "&&"),
            (Name.Builtin, "cd"),
        ],
        "cd || non-existance-cmd": [
            (Name.Builtin, "cd"),
            (Operator, "||"),
            (Error, "non-existance-cmd"),
        ],
    },
    "nested": {
        'echo @("hello")': [
            (Name.Builtin, "echo"),
            (Keyword, "@"),
            (Punctuation, "("),
            (String.Double, "hello"),
            (Punctuation, ")"),
        ],
        "print($(cd))": [
            (Name.Builtin, "print"),
            (Punctuation, "("),
            (Keyword, "$"),
            (Punctuation, "("),
            (Name.Builtin, "cd"),
            (Punctuation, ")"),
            (Punctuation, ")"),
            (Text, "\n"),
        ],
        r'print(![echo "])\""])': [
            (Name.Builtin, "print"),
            (Punctuation, "("),
            (Keyword, "!"),
            (Punctuation, "["),
            (Name.Builtin, "echo"),
            (Text, " "),
            (Literal.String.Double, '"])\\""'),
            (Punctuation, "]"),
            (Punctuation, ")"),
            (Text, "\n"),
        ],
    },
    "subproc-args": {
        "cd 192.168.0.1": [
            (Text, "192.168.0.1"),
        ],
    },
    "backtick": {
        r"echo g`.*\w+`": [
            (String.Affix, "g"),
            (String.Backtick, "`"),
            (String.Regex, "."),
            (String.Regex, "*"),
            (String.Escape, r"\w"),
        ],
    },
    "macro": {
        r"g!(42, *, 65)": [
            (Name, "g"),
            (Keyword, "!"),
            (Punctuation, "("),
            (Number.Integer, "42"),
        ],
        r"echo! hello world": [
            (Name.Builtin, "echo"),
            (Keyword, "!"),
            (String, "hello world"),
        ],
        r"bash -c ! export var=42; echo $var": [
            (Name.Builtin, "bash"),
            (Text, "-c"),
            (Keyword, "!"),
            (String, "export var=42; echo $var"),
        ],
    },
}


def _convert_cases():
    for title, input_dict in _cases.items():
        for idx, item in enumerate(input_dict.items()):
            yield pytest.param(*item, id=f"{title}-{idx}")


@pytest.mark.parametrize("inp, expected", list(_convert_cases()))
@skip_if_on_windows
def test_xonsh_lexer(inp, expected, check_token):
    check_token(inp, expected)


# can't seem to get thie test to import pyghooks and define on_lscolors_change handler like live code does.
# so we declare the event handler directly here.
@pytest.fixture
def events_fxt():
    return EventManager()


@pytest.fixture
def xonsh_builtins_ls_colors(xession, events_fxt):
    xession.shell = DummyShell()  # because load_command_cache zaps it.
    xession.shell.shell_type = "prompt_toolkit"
    lsc = LsColors(LsColors.default_settings)
    xession.env["LS_COLORS"] = lsc  # establish LS_COLORS before style.
    xession.shell.shell.styler = XonshStyle()  # default style

    events.on_lscolors_change(on_lscolors_change)

    yield xession


@skip_if_on_windows
def test_path(tmpdir, xonsh_builtins_ls_colors, check_token):
    test_dir = str(tmpdir.mkdir("xonsh-test-highlight-path"))
    check_token(f"cd {test_dir}", [(Name.Builtin, "cd"), (Color.BOLD_BLUE, test_dir)])
    check_token(
        f"cd {test_dir}-xxx",
        [(Name.Builtin, "cd"), (Text, f"{test_dir}-xxx")],
    )
    check_token(f"cd X={test_dir}", [(Color.BOLD_BLUE, test_dir)])

    with xonsh_builtins_ls_colors.env.swap(AUTO_CD=True):
        check_token(test_dir, [(Name.Constant, test_dir)])


@skip_if_on_windows
def test_color_on_lscolors_change(tmpdir, xonsh_builtins_ls_colors, check_token):
    """Verify colorizer returns Token.Text if file type not defined in LS_COLORS"""

    lsc = xonsh_builtins_ls_colors.env["LS_COLORS"]
    test_dir = str(tmpdir.mkdir("xonsh-test-highlight-path"))

    lsc["di"] = ("GREEN",)

    check_token(f"cd {test_dir}", [(Name.Builtin, "cd"), (Color.GREEN, test_dir)])

    del lsc["di"]

    check_token(f"cd {test_dir}", [(Name.Builtin, "cd"), (Text, test_dir)])



"""functions to update rc files"""
import os
import typing as tp

RENDERERS: tp.List[tp.Callable] = []


def renderer(f):
    """Adds decorated function to renderers list."""
    RENDERERS.append(f)


@renderer
def prompt(config):
    prompt = config.get("prompt")
    if prompt:
        yield f"$PROMPT = {prompt!r}"


@renderer
def colors(config):
    style = config.get("color")
    if style:
        yield f"$XONSH_COLOR_STYLE = {style!r}"


@renderer
def xontribs(config):
    xtribs = config.get("xontribs")
    if xtribs:
        yield "xontrib load " + " ".join(xtribs)


def config_to_xonsh(
    config, prefix="# XONSH WEBCONFIG START", suffix="# XONSH WEBCONFIG END"
):
    """Turns config dict into xonsh code (str)."""
    lines = [prefix]
    for func in RENDERERS:
        lines.extend(func(config))
    lines.append(suffix)
    return "\n".join(lines)


def insert_into_xonshrc(
    config,
    xonshrc="~/.xonshrc",
    prefix="# XONSH WEBCONFIG START",
    suffix="# XONSH WEBCONFIG END",
):
    """Places a config dict into the xonshrc."""
    # get current contents
    fname = os.path.expanduser(xonshrc)
    if os.path.isfile(fname):
        with open(fname) as f:
            s = f.read()
        before, _, s = s.partition(prefix)
        _, _, after = s.partition(suffix)
    else:
        before = after = ""
        dname = os.path.dirname(fname)
        if dname:
            os.makedirs(dname, exist_ok=True)
    # compute new values
    new = config_to_xonsh(config, prefix=prefix, suffix=suffix)
    # write out the file
    with open(fname, "w", encoding="utf-8") as f:
        f.write(before + new + after)
    return fname



"""
Command abbreviations.

This expands input words from `abbrevs` disctionary as you type.
Adds ``abbrevs`` dictionary to hold user-defined "command abbreviations.
The dictionary is searched as you type the matching words are replaced
at the command line by the corresponding dictionary contents once you hit
'Space' or 'Return' key.

For instance a frequently used command such as ``git status`` can be abbreviated to ``gst`` as follows::

    $ xontrib load abbrevs
    $ abbrevs['gst'] = 'git status'
    $ gst # Once you hit <space> or <return> 'gst' gets expanded to 'git status'.

one can set a callback function that receives current buffer and word to customize the expanded word based on context

.. code-block:: python

    $ abbrevs['ps'] = lambda buffer, word: "procs" if buffer.text.startswith(word) else word


It is also possible to set the cursor position after expansion with,

    $ abbrevs['gp'] = "git push <edit> --force"
"""

import builtins
import typing as tp

from prompt_toolkit.buffer import Buffer
from prompt_toolkit.filters import IsMultiline, completion_is_selected
from prompt_toolkit.keys import Keys

from xonsh.built_ins import DynamicAccessProxy, XonshSession
from xonsh.tools import check_for_partial_string

__all__ = ()


if tp.TYPE_CHECKING:

    class AbbrCallType(tp.Protocol):
        def __call__(self, word: str, buffer: Buffer) -> str:
            ...

    AbbrValType = tp.Union[str, AbbrCallType]

abbrevs: "dict[str, AbbrValType]" = dict()


class _LastExpanded(tp.NamedTuple):
    word: str
    expanded: str


class Abbreviation:
    """A container class to handle state related to abbreviating keywords"""

    last_expanded: tp.Optional[_LastExpanded] = None

    def expand(self, buffer: Buffer) -> bool:
        """expand the given abbr text. Return true if cursor position changed."""
        if not abbrevs:
            return False
        document = buffer.document
        word = document.get_word_before_cursor(WORD=True)
        if word in abbrevs.keys():
            partial = document.text[: document.cursor_position]
            startix, endix, quote = check_for_partial_string(partial)
            if startix is not None and endix is None:
                return False
            text = get_abbreviated(word, buffer)

            buffer.delete_before_cursor(count=len(word))
            buffer.insert_text(text)

            self.last_expanded = _LastExpanded(word, text)
            if EDIT_SYMBOL in text:
                set_cursor_position(buffer, text)
                return True
        return False

    def revert(self, buffer) -> bool:
        if self.last_expanded is None:
            return False
        document = buffer.document
        expansion = self.last_expanded.expanded + " "
        if not document.text_before_cursor.endswith(expansion):
            return False
        buffer.delete_before_cursor(count=len(expansion))
        buffer.insert_text(self.last_expanded.word)
        self.last_expanded = None
        return True


EDIT_SYMBOL = "<edit>"


def get_abbreviated(key: str, buffer) -> str:
    abbr = abbrevs[key]
    if callable(abbr):
        text = abbr(buffer=buffer, word=key)
    else:
        text = abbr
    return text


def set_cursor_position(buffer, expanded: str) -> None:
    pos = expanded.rfind(EDIT_SYMBOL)
    if pos == -1:
        return
    buffer.cursor_position = buffer.cursor_position - (len(expanded) - pos)
    buffer.delete(len(EDIT_SYMBOL))


def custom_keybindings(bindings, **kw):

    from prompt_toolkit.filters import EmacsInsertMode, ViInsertMode

    from xonsh.ptk_shell.key_bindings import carriage_return

    handler = bindings.add
    insert_mode = ViInsertMode() | EmacsInsertMode()
    abbrev = Abbreviation()

    @handler(" ", filter=IsMultiline() & insert_mode)
    def handle_space(event):
        buffer = event.app.current_buffer

        add_space = True
        if not abbrev.revert(buffer):
            position_changed = abbrev.expand(buffer)
            if position_changed:
                add_space = False
        if add_space:
            buffer.insert_text(" ")

    @handler(
        Keys.ControlJ, filter=IsMultiline() & insert_mode & ~completion_is_selected
    )
    @handler(
        Keys.ControlM, filter=IsMultiline() & insert_mode & ~completion_is_selected
    )
    def multiline_carriage_return(event):
        buffer = event.app.current_buffer
        current_char = buffer.document.current_char
        if not current_char or current_char.isspace():
            abbrev.expand(buffer)
        carriage_return(buffer, event.cli)


def _load_xontrib_(xsh: XonshSession, **_):
    xsh.builtins.events.on_ptk_create(custom_keybindings)
    # XSH.builtins is a namespace and extendable
    xsh.builtins.abbrevs = abbrevs
    proxy = DynamicAccessProxy("abbrevs", "__xonsh__.builtins.abbrevs")
    builtins.abbrevs = proxy  # type: ignore

    return {"abbrevs": abbrevs}



import pytest

from xonsh import jobs


@pytest.mark.parametrize(
    "args, prefix, exp",
    [
        (
            "disown",
            "-",
            {"-h", "--help", "-c", "--continue"},
        ),
        (
            "disown",
            "",
            {"1", "2"},
        ),
    ],
)
def test_disown_completion(
    args, prefix, exp, xsh_with_aliases, monkeypatch, check_completer
):
    job = {
        "cmds": (["git-cola", "2>", "/dev/null"], "&"),
        "pids": [37078],
        "bg": True,
        "pgrp": None,
        "started": 1630158319.697764,
        "status": "running",
    }
    all_jobs = {1: job, 2: job}

    monkeypatch.setattr(xsh_with_aliases, "all_jobs", all_jobs)
    monkeypatch.setattr(jobs, "tasks", [2, 1])
    assert check_completer(args, prefix=prefix) == exp



"""Implements the xonsh parser for Python v3.6."""
import xonsh.ast as ast
from xonsh.parsers.base import BaseParser, lopen_loc, store_ctx


class Parser(BaseParser):
    """A Python v3.6 compliant parser for the xonsh language."""

    def __init__(
        self,
        yacc_optimize=True,
        yacc_table="xonsh.parser_table",
        yacc_debug=False,
        outputdir=None,
    ):
        """
        Parameters
        ----------
        yacc_optimize : bool, optional
            Set to false when unstable and true when parser is stable.
        yacc_table : str, optional
            Parser module used when optimized.
        yacc_debug : debug, optional
            Dumps extra debug info.
        outputdir : str or None, optional
            The directory to place generated tables within.
        """
        # Rule creation and modification *must* take place before super()
        tok_rules = ["await", "async"]
        for rule in tok_rules:
            self._tok_rule(rule)
        super().__init__(
            yacc_optimize=yacc_optimize,
            yacc_table=yacc_table,
            yacc_debug=yacc_debug,
            outputdir=outputdir,
        )

    def p_classdef_or_funcdef(self, p):
        """
        classdef_or_funcdef : classdef
                            | funcdef
                            | async_funcdef
        """
        p[0] = p[1]

    def p_async_funcdef(self, p):
        """async_funcdef : async_tok funcdef"""
        p1, f = p[1], p[2][0]
        p[0] = [ast.AsyncFunctionDef(**f.__dict__)]
        p[0][0]._async_tok = p1

    def p_async_compound_stmt(self, p):
        """compound_stmt : async_stmt"""
        p[0] = p[1]

    def p_async_for_stmt(self, p):
        """async_for_stmt : ASYNC for_stmt"""
        f = p[2][0]
        p[0] = [ast.AsyncFor(**f.__dict__)]

    def p_async_with_stmt(self, p):
        """async_with_stmt : ASYNC with_stmt"""
        w = p[2][0]
        p[0] = [ast.AsyncWith(**w.__dict__)]

    def p_atom_expr_await(self, p):
        """atom_expr : await_tok atom trailer_list_opt"""
        p0 = self.apply_trailers(p[2], p[3])
        p1 = p[1]
        p0 = ast.Await(value=p0, ctx=ast.Load(), lineno=p1.lineno, col_offset=p1.lexpos)
        p[0] = p0

    def p_async_stmt(self, p):
        """
        async_stmt : async_funcdef
                   | async_with_stmt
                   | async_for_stmt
        """
        p[0] = p[1]

    def p_item_test(self, p):
        """item : test COLON test"""
        p[0] = [p[1], p[3]]

    def p_item_pow(self, p):
        """item : POW expr"""
        p[0] = [None, p[2]]

    def _set_arg(self, args, arg, ensure_kw=False):
        if isinstance(arg, ast.keyword):
            args["keywords"].append(arg)
        elif ensure_kw:
            args["keywords"].append(ast.keyword(arg=None, value=arg))
        else:
            args["args"].append(arg)

    def p_arglist_single(self, p):
        """arglist : argument comma_opt"""
        p0 = {"args": [], "keywords": []}
        self._set_arg(p0, p[1])
        p[0] = p0

    def p_arglist_many(self, p):
        """arglist : argument comma_argument_list comma_opt"""
        p0 = {"args": [], "keywords": []}
        self._set_arg(p0, p[1])
        for arg in p[2]:
            self._set_arg(p0, arg)
        p[0] = p0

    # Argument rules
    # "test '=' test" is really "keyword '=' test", but we have no such token.
    # These need to be in a single rule to avoid grammar that is ambiguous
    # to our LL(1) parser. Even though 'test' includes '*expr' in star_expr,
    # we explicitly match '*' here, too, to give it proper precedence.
    # Illegal combinations and orderings are blocked in ast.c:
    # multiple (test comp_for) arguments are blocked; keyword unpackings
    # that precede iterable unpackings are blocked; etc.
    def p_argument_test_or_star(self, p):
        """argument : test_or_star_expr"""
        p[0] = p[1]

    def p_argument_kwargs(self, p):
        """argument : POW test"""
        p2 = p[2]
        p[0] = ast.keyword(
            arg=None, value=p2, lineno=p2.lineno, col_offset=p2.col_offset
        )

    def p_argument_args(self, p):
        """argument : TIMES test"""
        p[0] = ast.Starred(value=p[2])

    def p_argument(self, p):
        """argument : test comp_for"""
        p1 = p[1]
        p[0] = ast.GeneratorExp(
            elt=p1, generators=p[2]["comps"], lineno=p1.lineno, col_offset=p1.col_offset
        )

    def p_argument_eq(self, p):
        """argument : test EQUALS test"""
        p3 = p[3]
        p[0] = ast.keyword(
            arg=p[1].id, value=p3, lineno=p3.lineno, col_offset=p3.col_offset
        )

    def p_comp_for(self, p):
        """comp_for : FOR exprlist IN or_test comp_iter_opt"""
        super().p_comp_for(p)
        # only difference with base should be the is_async=0
        p[0]["comps"][0].is_async = 0

    def p_expr_stmt_annassign(self, p):
        """expr_stmt : testlist_star_expr COLON test EQUALS test
        | testlist_star_expr COLON test
        """
        p1 = p[1][0]
        lineno, col = lopen_loc(p1)
        if len(p[1]) > 1 or not isinstance(
            p1, (ast.Name, ast.Attribute, ast.Subscript)
        ):
            loc = self.currloc(lineno, col)
            self._set_error("only single target can be annotated", loc)
        store_ctx(p1)
        p[0] = ast.AnnAssign(
            target=p1,
            annotation=p[3],
            value=p[5] if len(p) >= 6 else None,
            simple=1,
            lineno=lineno,
            col_offset=col,
        )



"""Testing for ``xonsh.shell.Shell``"""
import os

from xonsh.history.dummy import DummyHistory
from xonsh.history.json import JsonHistory
from xonsh.history.sqlite import SqliteHistory
from xonsh.shell import Shell


def test_shell_with_json_history(xession, xonsh_execer, tmpdir_factory):
    """
    Check that shell successfully load JSON history from file.
    """
    tempdir = str(tmpdir_factory.mktemp("history"))

    history_file = os.path.join(tempdir, "history.json")
    h = JsonHistory(filename=history_file)
    h.append(
        {
            "inp": "echo Hello world 1\n",
            "rtn": 0,
            "ts": [1615887820.7329783, 1615887820.7513437],
        }
    )
    h.append(
        {
            "inp": "echo Hello world 2\n",
            "rtn": 0,
            "ts": [1615887820.7329783, 1615887820.7513437],
        }
    )
    h.flush()

    xession.env.update(
        dict(
            XONSH_DATA_DIR=tempdir,
            XONSH_INTERACTIVE=True,
            XONSH_HISTORY_BACKEND="json",
            XONSH_HISTORY_FILE=history_file,
            # XONSH_DEBUG=1  # to show errors
        )
    )

    Shell(xonsh_execer, shell_type="none")

    assert len([i for i in xession.history.all_items()]) == 2


def test_shell_with_sqlite_history(xession, xonsh_execer, tmpdir_factory):
    """
    Check that shell successfully load SQLite history from file.
    """
    tempdir = str(tmpdir_factory.mktemp("history"))

    history_file = os.path.join(tempdir, "history.db")
    h = SqliteHistory(filename=history_file)
    h.append(
        {
            "inp": "echo Hello world 1\n",
            "rtn": 0,
            "ts": [1615887820.7329783, 1615887820.7513437],
        }
    )
    h.append(
        {
            "inp": "echo Hello world 2\n",
            "rtn": 0,
            "ts": [1615887820.7329783, 1615887820.7513437],
        }
    )
    h.flush()

    xession.env.update(
        dict(
            XONSH_DATA_DIR=tempdir,
            XONSH_INTERACTIVE=True,
            XONSH_HISTORY_BACKEND="sqlite",
            XONSH_HISTORY_FILE=history_file,
            # XONSH_DEBUG=1  # to show errors
        )
    )

    Shell(xonsh_execer, shell_type="none")

    assert len([i for i in xession.history.all_items()]) == 2


def test_shell_with_dummy_history_in_not_interactive(xession, xonsh_execer):
    """
    Check that shell use Dummy history in not interactive mode.
    """
    xession.env["XONSH_INTERACTIVE"] = False
    xession.history = None
    Shell(xonsh_execer, shell_type="none")
    assert isinstance(xession.history, DummyHistory)



from unittest.mock import Mock

import pytest

from xonsh.completer import Completer
from xonsh.completers.commands import complete_command, complete_skipper
from xonsh.parsers.completion_context import (
    CommandArg,
    CommandContext,
    CompletionContext,
)
from xonsh.pytest.tools import ON_WINDOWS, completions_from_result, skip_if_on_windows


@pytest.fixture(autouse=True)
def xs_orig_commands_cache(xession):
    pass


def test_complete_command(completion_context_parse):
    if ON_WINDOWS:
        command = "dir.exe"
    else:
        command = "grep"

    comps = complete_command(
        completion_context_parse(command[:-1], len(command) - 1).command
    )
    assert command in set(map(str, comps))


@skip_if_on_windows
def test_skipper_command(completion_context_parse):
    assert "grep" in completions_from_result(
        complete_skipper(completion_context_parse("sudo gre", 8))
    )


@skip_if_on_windows
def test_skipper_arg(completion_context_parse, xession, monkeypatch):
    monkeypatch.setattr(xession.shell.shell, "completer", Completer(), raising=False)
    bash_completer_mock = Mock()
    monkeypatch.setattr(xession, "completers", {"bash": bash_completer_mock})

    bash_completer_mock.return_value = {"--count "}

    assert "--count " in completions_from_result(
        complete_skipper(completion_context_parse("sudo grep --coun", 16))
    )

    call_args = bash_completer_mock.call_args[0]
    assert len(call_args) == 1

    context = call_args[0]
    assert isinstance(context, CompletionContext)
    assert context.command == CommandContext(
        args=(CommandArg("grep"),), arg_index=1, prefix="--coun"
    )



import re

import pytest

from xonsh.procs.specs import cmds_to_specs
from xonsh.tracer import tracermain


def test_tracer_help(capsys, xsh_with_aliases):
    """verify can invoke it, and usage knows about all the options"""
    spec = cmds_to_specs([("trace", "-h")], captured="stdout")[0]
    with pytest.raises(SystemExit):
        tracermain(["-h"], spec=spec)
    capout = capsys.readouterr().out
    pat = re.compile(r"^usage:\s*trace[^\n]*{([\w,-]+)}", re.MULTILINE)
    m = pat.match(capout)
    assert m[1]
    verbs = {v.strip().lower() for v in m[1].split(",")}
    assert verbs == {"rm", "start", "add", "on", "off", "del", "color", "stop", "ls"}



"""Testing xonsh import hooks"""
import os
from importlib import import_module

import pytest

from xonsh import imphooks


@pytest.fixture(autouse=True)
def imp_env(xession):
    xession.env.update({"PATH": [], "PATHEXT": []})
    imphooks.install_import_hooks(xession.execer)
    yield


def test_import():
    import sample

    assert "hello mom jawaka\n" == sample.x


def test_import_empty():
    from xpack import empty_xsh

    assert empty_xsh


def test_absolute_import():
    from xpack import sample

    assert "hello mom jawaka\n" == sample.x


def test_relative_import():
    from xpack import relimp

    assert "hello mom jawaka\n" == relimp.sample.x
    assert "hello mom jawaka\ndark chest of wonders" == relimp.y


def test_sub_import():
    from xpack.sub import sample

    assert "hello mom jawaka\n" == sample.x


TEST_DIR = os.path.dirname(__file__)


def test_module_dunder_file_attribute():
    import sample

    exp = os.path.join(TEST_DIR, "sample.xsh")
    assert os.path.abspath(sample.__file__) == exp


def test_module_dunder_file_attribute_sub():
    from xpack.sub import sample

    exp = os.path.join(TEST_DIR, "xpack", "sub", "sample.xsh")
    assert os.path.abspath(sample.__file__) == exp


def test_get_source():
    mod = import_module("sample")
    loader = mod.__loader__
    source = loader.get_source("sample")
    with open(os.path.join(TEST_DIR, "sample.xsh")) as srcfile:
        assert source == srcfile.read()



"""Test module xonsh/cli_utils.py"""

from xonsh import cli_utils


def func_with_doc(param: str, multi: str, optional=False):
    """func doc
    multi-line

    Parameters
    ----------
    param
        param doc
    multi
        param doc
        multi line
    optional : -o, --opt
        an optional parameter with flags defined in description

    Returns
    -------
    str
        return doc
    """
    return param + multi, optional


def test_get_doc_param():
    doc = cli_utils.NumpyDoc(func_with_doc)
    assert doc.description.splitlines() == [
        "func doc",
        "multi-line",
    ]
    assert doc.epilog.splitlines() == [
        "Returns",
        "-------",
        "str",
        "    return doc",
    ]
    assert doc.params["param"].splitlines() == [
        "param doc",
    ]
    assert doc.params["multi"].splitlines() == [
        "param doc",
        "multi line",
    ]
    assert doc.flags == {"optional": ["-o", "--opt"]}


def test_generated_parser():
    from xonsh.completers._aliases import CompleterAlias

    alias = CompleterAlias()

    assert alias.parser.description

    positionals = alias.parser._get_positional_actions()
    add_cmd = positionals[0].choices["add"]
    assert "Add a new completer" in add_cmd.description
    assert (
        alias.parser.format_usage()
        == "usage: completer [-h] {add,remove,rm,list,ls,complete} ...\n"
    )
    assert add_cmd.format_usage() == "usage: completer add [-h] name func [pos]\n"


def test_parser_hooking():
    from xonsh.history.main import HistoryAlias

    alias = HistoryAlias()

    parser, _ = cli_utils.ArgparseCompleter.get_parser(alias.parser, ["show"])
    assert parser._get_positional_actions()[0].choices == (
        "session",
        "xonsh",
        "all",
        "zsh",
        "bash",
    )



"""Tests the xonsh.procs.specs"""
import itertools
import sys
from subprocess import Popen

import pytest

from xonsh.procs.posix import PopenThread
from xonsh.procs.proxies import STDOUT_DISPATCHER, ProcProxy, ProcProxyThread
from xonsh.procs.specs import cmds_to_specs, run_subproc
from xonsh.pytest.tools import skip_if_on_windows


@skip_if_on_windows
def test_cmds_to_specs_thread_subproc(xession):
    env = xession.env
    cmds = [["pwd"]]

    # XONSH_CAPTURE_ALWAYS=False should disable interactive threaded subprocs
    env["XONSH_CAPTURE_ALWAYS"] = False
    env["THREAD_SUBPROCS"] = True
    specs = cmds_to_specs(cmds, captured="hiddenobject")
    assert specs[0].cls is Popen

    # Now for the other situations
    env["XONSH_CAPTURE_ALWAYS"] = True

    # First check that threadable subprocs become threadable
    env["THREAD_SUBPROCS"] = True
    specs = cmds_to_specs(cmds, captured="hiddenobject")
    assert specs[0].cls is PopenThread
    # turn off threading and check we use Popen
    env["THREAD_SUBPROCS"] = False
    specs = cmds_to_specs(cmds, captured="hiddenobject")
    assert specs[0].cls is Popen

    # now check the threadbility of callable aliases
    cmds = [[lambda: "Keras Selyrian"]]
    # check that threadable alias become threadable
    env["THREAD_SUBPROCS"] = True
    specs = cmds_to_specs(cmds, captured="hiddenobject")
    assert specs[0].cls is ProcProxyThread
    # turn off threading and check we use ProcProxy
    env["THREAD_SUBPROCS"] = False
    specs = cmds_to_specs(cmds, captured="hiddenobject")
    assert specs[0].cls is ProcProxy


@pytest.mark.parametrize("thread_subprocs", [True, False])
def test_cmds_to_specs_capture_stdout_not_stderr(thread_subprocs, xonsh_session):
    env = xonsh_session.env
    cmds = (["ls", "/root"],)

    env["THREAD_SUBPROCS"] = thread_subprocs

    specs = cmds_to_specs(cmds, captured="stdout")
    assert specs[0].stdout is not None
    assert specs[0].stderr is None


@skip_if_on_windows
@pytest.mark.parametrize("pipe", (True, False))
@pytest.mark.parametrize("alias_type", (None, "func", "exec", "simple"))
@pytest.mark.parametrize(
    "thread_subprocs, capture_always", list(itertools.product((True, False), repeat=2))
)
@pytest.mark.flaky(reruns=5, reruns_delay=2)
def test_capture_always(
    capfd, thread_subprocs, capture_always, alias_type, pipe, monkeypatch, xonsh_session
):
    if not thread_subprocs and alias_type in ["func", "exec"]:
        if pipe:
            return pytest.skip("https://github.com/xonsh/xonsh/issues/4443")
        else:
            return pytest.skip("https://github.com/xonsh/xonsh/issues/4444")

    env = xonsh_session.env
    exp = "HELLO\nBYE\n"
    cmds = [["echo", "-n", exp]]
    if pipe:
        exp = exp.splitlines()[1] + "\n"  # second line
        cmds += ["|", ["grep", "--color=never", exp.strip()]]

    if alias_type:
        first_cmd = cmds[0]
        # Enable capfd for function aliases:
        monkeypatch.setattr(STDOUT_DISPATCHER, "default", sys.stdout)
        if alias_type == "func":
            xonsh_session.aliases["tst"] = (
                lambda: run_subproc([first_cmd], "hiddenobject") and None
            )  # Don't return a value
        elif alias_type == "exec":
            first_cmd = " ".join(repr(arg) for arg in first_cmd)
            xonsh_session.aliases["tst"] = f"![{first_cmd}]"
        else:
            # alias_type == "simple"
            xonsh_session.aliases["tst"] = first_cmd

        cmds[0] = ["tst"]

    env["THREAD_SUBPROCS"] = thread_subprocs
    env["XONSH_CAPTURE_ALWAYS"] = capture_always

    hidden = run_subproc(cmds, "hiddenobject")  # ![]
    # Check that interactive subprocs are always printed
    assert exp in capfd.readouterr().out

    if capture_always and thread_subprocs:
        # Check that the interactive output was captured
        assert hidden.out == exp
    else:
        # without THREAD_SUBPROCS capturing in ![] isn't possible
        assert not hidden.out

    # Explicitly captured commands are always captured
    hidden = run_subproc(cmds, "object")  # !()
    hidden.end()
    if thread_subprocs:
        assert exp not in capfd.readouterr().out
        assert hidden.out == exp
    else:
        # for some reason THREAD_SUBPROCS=False fails to capture in `!()` but still succeeds in `$()`
        assert exp in capfd.readouterr().out
        assert not hidden.out

    output = run_subproc(cmds, "stdout")  # $()
    assert exp not in capfd.readouterr().out
    assert output == exp

    # Explicitly non-captured commands are never captured (/always printed)
    run_subproc(cmds, captured=False)  # $[]
    assert exp in capfd.readouterr().out


@skip_if_on_windows
@pytest.mark.parametrize(
    "captured, exp_is_none",
    [
        ("object", False),
        ("stdout", True),
        ("hiddenobject", False),
        (False, True),
    ],
)
def test_run_subproc_background(captured, exp_is_none):

    cmds = (["echo", "hello"], "&")
    return_val = run_subproc(cmds, captured)
    assert (return_val is None) == exp_is_none


@pytest.mark.parametrize("thread_subprocs", [False, True])
def test_callable_alias_cls(thread_subprocs, xession):
    class Cls:
        def __call__(self, *args, **kwargs):
            print(args, kwargs)

    obj = Cls()
    xession.aliases["tst"] = obj

    env = xession.env
    cmds = (["tst", "/root"],)

    env["THREAD_SUBPROCS"] = thread_subprocs

    spec = cmds_to_specs(cmds, captured="stdout")[0]
    proc = spec.run()
    assert proc.f == obj



"""Python virtual environment manager for xonsh."""
import os.path
import subprocess
import tempfile
import typing as tp
from pathlib import Path

import xonsh.cli_utils as xcli
import xontrib.voxapi as voxapi
from xonsh.built_ins import XSH, XonshSession
from xonsh.dirstack import pushd_fn
from xonsh.platform import ON_WINDOWS
from xonsh.tools import XonshError

__all__ = ()


def venv_names_completer(command, alias: "VoxHandler", **_):
    envs = alias.vox.keys()
    from xonsh.completers.path import complete_dir

    yield from envs

    paths, _ = complete_dir(command)
    yield from paths


def py_interpreter_path_completer(xsh, **_):
    for _, (path, is_alias) in xsh.commands_cache.all_commands.items():
        if not is_alias and ("/python" in path or "/pypy" in path):
            yield path


_venv_option = xcli.Annotated[
    tp.Optional[str],
    xcli.Arg(metavar="ENV", nargs="?", completer=venv_names_completer),
]


class VoxHandler(xcli.ArgParserAlias):
    """Vox is a virtual environment manager for xonsh."""

    def build(self):
        """lazily called during dispatch"""
        self.vox = voxapi.Vox()
        parser = self.create_parser(prog="vox")

        parser.add_command(self.new, aliases=["create"])
        parser.add_command(self.activate, aliases=["workon", "enter"])
        parser.add_command(self.deactivate, aliases=["exit"])
        parser.add_command(self.list, aliases=["ls"])
        parser.add_command(self.remove, aliases=["rm", "delete", "del"])
        parser.add_command(self.info)
        parser.add_command(self.runin)
        parser.add_command(self.runin_all)
        parser.add_command(self.toggle_ssp)
        parser.add_command(self.wipe)
        parser.add_command(self.project_set)
        parser.add_command(self.project_get)
        parser.add_command(self.upgrade)

        return parser

    def hook_pre_add_argument(self, param: str, func, flags, kwargs):
        if func.__name__ in {"new", "upgrade"}:
            if ON_WINDOWS and param == "symlinks":
                # copies by default on windows
                kwargs["default"] = False
                kwargs["action"] = "store_true"
                kwargs["help"] = "Try to use symlinks rather than copies"
                flags = ["--symlinks"]
        return flags, kwargs

    def hook_post_add_argument(self, action, param: str, **_):
        if param == "interpreter":
            action.completer = py_interpreter_path_completer

    def new(
        self,
        name: xcli.Annotated[str, xcli.Arg(metavar="ENV")],
        interpreter: "str|None" = None,
        system_site_packages=False,
        symlinks=True,
        without_pip=False,
        activate=False,
        temporary=False,
        packages: xcli.Annotated[tp.Sequence[str], xcli.Arg(nargs="*")] = (),
        requirements: xcli.Annotated[tp.Sequence[str], xcli.Arg(action="append")] = (),
        link_project_dir=False,
        prompt: "str|None" = None,
    ):
        """Create a virtual environment in $VIRTUALENV_HOME with python3's ``venv``.

        Parameters
        ----------
        name : str
            Virtual environment name
        interpreter: -p, --interpreter
            Python interpreter used to create the virtual environment.
            Can be configured via the $VOX_DEFAULT_INTERPRETER environment variable.
        system_site_packages : --system-site-packages, --ssp
            If True, the system (global) site-packages dir is available to
            created environments.
        symlinks : --copies
            Try to use copies rather than symlinks.
        without_pip : --without-pip, --wp
            Skips installing or upgrading pip in the virtual environment
        activate : -a, --activate
            Activate the newly created virtual environment.
        temporary: -t, --temp
            Create the virtualenv under a temporary directory.
        packages: -i, --install
            Install one or more packages (by repeating the option) after the environment is created using pip
        requirements: -r, --requirements
            The argument value is passed to ``pip -r`` to be installed.
        link_project_dir: -l, --link, --link-project
            Associate the current directory with the new environment.
        prompt: --prompt
            Provides an alternative prompt prefix for this environment.
        """

        self.out("Creating environment...")

        if temporary:
            path = tempfile.mkdtemp(prefix=f"vox-env-{name}")
            name = os.path.join(path, name)

        self.vox.create(
            name,
            system_site_packages=system_site_packages,
            symlinks=symlinks,
            with_pip=(not without_pip),
            interpreter=interpreter,
            prompt=prompt,
        )
        if link_project_dir:
            self.project_set(name)

        if packages:
            self.runin(name, ["pip", "install", *packages])

        if requirements:

            def _generate_args():
                for req in requirements:
                    yield "-r"
                    yield req

            self.runin(name, ["pip", "install"] + list(_generate_args()))

        if activate:
            self.activate(name)
            self.out(f"Environment {name!r} created and activated.\n")
        else:
            self.out(
                f'Environment {name!r} created. Activate it with "vox activate {name}".\n'
            )

    def activate(
        self,
        name: _venv_option = None,
        no_cd=False,
    ):
        """Activate a virtual environment.

        Parameters
        ----------
        name
            The environment to activate.
            ENV can be either a name from the venvs shown by ``vox list``
            or the path to an arbitrary venv
        no_cd: -n, --no-cd
            Do not change current working directory even if a project path is associated with ENV.
        """

        if name is None:
            return self.list()

        try:
            self.vox.activate(name)
        except KeyError:
            raise self.Error(
                f'This environment doesn\'t exist. Create it with "vox new {name}"',
            )

        self.out(f'Activated "{name}".\n')
        if not no_cd:
            project_dir = self._get_project_dir(name)
            if project_dir:
                pushd_fn(project_dir)

    def deactivate(self, remove=False, force=False):
        """Deactivate the active virtual environment.

        Parameters
        ----------
        remove: -r, --remove
            Remove the virtual environment after leaving it.
        force: -f, --force-removal
            Remove the virtual environment without prompt
        """

        if self.vox.active() is None:
            raise self.Error(
                'No environment currently active. Activate one with "vox activate".\n',
            )
        env_name = self.vox.deactivate()
        if remove:
            self.vox.force_removals = force
            del self.vox[env_name]
            self.out(f'Environment "{env_name}" deactivated and removed.\n')
        else:
            self.out(f'Environment "{env_name}" deactivated.\n')

    def list(self):
        """List available virtual environments."""

        try:
            envs = sorted(self.vox.keys())
        except PermissionError:
            raise self.Error("No permissions on VIRTUALENV_HOME")

        if not envs:
            raise self.Error(
                'No environments available. Create one with "vox new".\n',
            )

        self.out("Available environments:")
        self.out("\n".join(envs))

    def remove(
        self,
        names: xcli.Annotated[
            tp.List[str],
            xcli.Arg(metavar="ENV", nargs="+", completer=venv_names_completer),
        ],
        force=False,
    ):
        """Remove virtual environments.

        Parameters
        ----------
        names
            The environments to remove. ENV can be either a name from the venvs shown by vox
            list or the path to an arbitrary venv
        force : -f, --force
            Delete virtualenv without prompt
        """
        self.vox.force_removals = force
        for name in names:
            try:
                del self.vox[name]
            except voxapi.EnvironmentInUse:
                raise self.Error(
                    f'The "{name}" environment is currently active. '
                    'In order to remove it, deactivate it first with "vox deactivate".\n',
                )
            except KeyError:
                raise self.Error(f'"{name}" environment doesn\'t exist.\n')
            else:
                self.out(f'Environment "{name}" removed.')
        self.out()

    def _in_venv(self, env_dir: str, command: str, *args, **kwargs):
        env = {**XSH.env.detype(), "VIRTUAL_ENV": env_dir}

        bin_path = os.path.join(env_dir, self.vox.sub_dirs[0])
        env["PATH"] = os.pathsep.join([bin_path, env["PATH"]])

        for key in ("PYTHONHOME", "__PYVENV_LAUNCHER__"):
            env.pop(key, None)

        try:
            return subprocess.check_call(
                [command] + list(args), shell=bool(ON_WINDOWS), env=env, **kwargs
            )
            # need to have shell=True on windows, otherwise the PYTHONPATH
            # won't inherit the PATH
        except OSError as e:
            if e.errno == 2:
                raise self.Error(f"Unable to find {command}")
            raise

    def runin(
        self,
        venv: xcli.Annotated[
            str,
            xcli.Arg(completer=venv_names_completer),
        ],
        args: xcli.Annotated[tp.Sequence[str], xcli.Arg(nargs="...")],
    ):
        """Run the command in the given environment

        Parameters
        ----------
        venv
            The environment to run the command for
        args
            The actual command to run

        Examples
        --------
          vox runin venv1 black --check-only
        """
        env_dir = self._get_env_dir(venv)
        if not args:
            raise self.Error("No command is passed")
        self._in_venv(env_dir, *args)

    def runin_all(
        self,
        args: xcli.Annotated[tp.Sequence[str], xcli.Arg(nargs="...")],
    ):
        """Run the command in all environments found under $VIRTUALENV_HOME

        Parameters
        ----------
        args
            The actual command to run with arguments
        """
        errors = False
        for env in self.vox:
            self.out("\n%s:" % env)
            try:
                self.runin(env, *args)
            except subprocess.CalledProcessError as e:
                errors = True
                self.err(e)
        self.parser.exit(errors)

    def _sitepackages_dir(self, venv_path: str):
        env_python = self.vox.get_binary_path("python", venv_path)
        if not os.path.exists(env_python):
            raise self.Error("no virtualenv active")

        return Path(
            subprocess.check_output(
                [
                    str(env_python),
                    "-c",
                    "import distutils; \
    print(distutils.sysconfig.get_python_lib())",
                ]
            ).decode()
        )

    def _get_env_dir(self, venv=None):
        venv = venv or ...
        try:
            env_dir = self.vox[venv].env
        except KeyError:
            # check whether the venv is a valid path to an environment
            if (
                isinstance(venv, str)
                and os.path.exists(venv)
                and os.path.exists(self.vox.get_binary_path("python", venv))
            ):
                return venv
            raise XonshError("No virtualenv is found")
        return env_dir

    def toggle_ssp(self):
        """Controls whether the active virtualenv will access the packages
        in the global Python site-packages directory."""
        # https://virtualenv.pypa.io/en/legacy/userguide.html#the-system-site-packages-option
        env_dir = self._get_env_dir()  # current
        site = self._sitepackages_dir(env_dir)
        ngsp_file = site.parent / "no-global-site-packages.txt"
        if ngsp_file.exists():
            ngsp_file.unlink()
            self.out("Enabled global site-packages")
        else:
            with ngsp_file.open("w"):
                self.out("Disabled global site-packages")

    def project_set(
        self,
        venv: _venv_option = None,
        project_path=None,
    ):
        """Bind an existing virtualenv to an existing project.

        Parameters
        ----------
        venv
            Name of the virtualenv, while the default being currently active venv.
        project_path
            Path to the project, while the default being current directory.
        """
        env_dir = self._get_env_dir(venv)  # current

        project = os.path.abspath(project_path or ".")
        if not os.path.exists(env_dir):
            raise self.Error(f"Environment '{env_dir}' doesn't exist.")
        if not os.path.isdir(project):
            raise self.Error(f"{project} does not exist")

        project_file = self._get_project_file()
        project_file.write_text(project)

    def _get_project_file(
        self,
        venv=None,
    ):
        env_dir = Path(self._get_env_dir(venv))  # current
        return env_dir / ".project"

    def _get_project_dir(self, venv=None):
        project_file = self._get_project_file(venv)
        if project_file.exists():
            project_dir = project_file.read_text()
            if os.path.exists(project_dir):
                return project_dir

    def project_get(self, venv: _venv_option = None):
        """Return a virtualenv's project directory.

        Parameters
        ----------
        venv
            Name of the virtualenv under $VIRTUALENV_HOME, while default being currently active venv.
        """
        project_dir = self._get_project_dir(venv)
        if project_dir:
            self.out(project_dir)
        else:
            project_file = self._get_project_file(venv)
            raise self.Error(
                f"Corrupted or outdated: {project_file}\nDirectory: {project_dir} doesn't exist."
            )

    def wipe(self, venv: _venv_option = None):
        """Remove all installed packages from the current (or supplied) env.

        Parameters
        ----------
        venv
            name of the venv. Defaults to currently active venv
        """
        env_dir = self._get_env_dir(venv)
        pip_bin = self.vox.get_binary_path("pip", env_dir)
        all_pkgs = set(
            subprocess.check_output([pip_bin, "freeze", "--local"])
            .decode()
            .splitlines()
        )
        pkgs = {p for p in all_pkgs if len(p.split("==")) == 2}
        ignored = sorted(all_pkgs - pkgs)
        to_remove = {p.split("==")[0] for p in pkgs}
        if to_remove:
            self.out("Ignoring:\n %s" % "\n ".join(ignored))
            self.out("Uninstalling packages:\n %s" % "\n ".join(to_remove))
            return subprocess.run([pip_bin, "uninstall", "-y", *to_remove])
        else:
            self.out("Nothing to remove")

    def info(self, venv: _venv_option = None):
        """Prints the path for the supplied env

        Parameters
        ----------
        venv
            name of the venv
        """
        self.out(self.vox[venv or ...])

    def upgrade(
        self,
        name: _venv_option = None,
        interpreter: "str|None" = None,
        symlinks=True,
        with_pip=False,
    ):
        """Upgrade the environment directory to use this version
        of Python, assuming Python has been upgraded in-place.

        WARNING: If a virtual environment was created with symlinks or without PIP, you must
        specify these options again on upgrade.

        Parameters
        ----------
        name
            Name or the path to the virtual environment
        interpreter: -p, --interpreter
            Python interpreter used to create the virtual environment.
            Can be configured via the $VOX_DEFAULT_INTERPRETER environment variable.
        symlinks : --copies
            Try to use copies rather than symlinks.
        with_pip : --without-pip, --wp
            Skips installing or upgrading pip in the virtual environment
        """
        venv = self.vox.upgrade(
            name or ..., symlinks=symlinks, with_pip=with_pip, interpreter=interpreter
        )
        self.out(venv)


def _load_xontrib_(xsh: XonshSession, **_):
    xsh.aliases["vox"] = VoxHandler(threadable=False)



"""Tests sample inputs to PTK multiline and checks parser response"""
from collections import namedtuple
from unittest.mock import MagicMock, patch

import pytest
from prompt_toolkit.application import Application
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.document import Document

from xonsh.tools import ON_WINDOWS

Context = namedtuple("Context", ["indent", "buffer", "accept", "cli", "cr"])


@pytest.fixture
def ctx(xession):
    """Context in which the ptk multiline functionality will be tested."""
    xession.env["INDENT"] = "    "
    from xonsh.ptk_shell.key_bindings import carriage_return

    ptk_buffer = Buffer()
    ptk_buffer.accept_action = MagicMock(name="accept")
    cli = MagicMock(name="cli", spec=Application)
    yield Context(
        indent="    ",
        buffer=ptk_buffer,
        accept=ptk_buffer.accept_action,
        cli=cli,
        cr=carriage_return,
    )


def test_colon_indent(ctx):
    document = Document("for i in range(5):")
    ctx.buffer.set_document(document)
    ctx.cr(ctx.buffer, ctx.cli)
    assert ctx.buffer.document.current_line == ctx.indent


def test_dedent(ctx):
    document = Document("\n" + ctx.indent + "pass")
    ctx.buffer.set_document(document)
    ctx.cr(ctx.buffer, ctx.cli)
    assert ctx.buffer.document.current_line == ""

    document = Document("\n" + 2 * ctx.indent + "continue")
    ctx.buffer.set_document(document)
    ctx.cr(ctx.buffer, ctx.cli)
    assert ctx.buffer.document.current_line == ctx.indent


def test_nodedent(ctx):
    """don't dedent if first line of ctx.buffer"""
    mock = MagicMock(return_value=True)
    with patch("xonsh.ptk_shell.key_bindings.can_compile", mock):
        document = Document("pass")
        ctx.buffer.set_document(document)
        ctx.cr(ctx.buffer, ctx.cli)
        assert ctx.accept.mock_calls is not None

    mock = MagicMock(return_value=True)
    with patch("xonsh.ptk_shell.key_bindings.can_compile", mock):
        document = Document(ctx.indent + "pass")
        ctx.buffer.set_document(document)
        ctx.cr(ctx.buffer, ctx.cli)
        assert ctx.accept.mock_calls is not None


def test_continuation_line(ctx):
    document = Document("\nsecond line")
    ctx.buffer.set_document(document)
    ctx.cr(ctx.buffer, ctx.cli)
    assert ctx.buffer.document.current_line == ""


def test_trailing_slash(ctx):
    mock = MagicMock(return_value=True)
    with patch("xonsh.ptk_shell.key_bindings.can_compile", mock):
        document = Document("this line will \\")
        ctx.buffer.set_document(document)
        ctx.cr(ctx.buffer, ctx.cli)
        if not ON_WINDOWS:
            assert ctx.buffer.document.current_line == ""
        else:
            assert ctx.accept.mock_calls is not None


def test_cant_compile_newline(ctx):
    mock = MagicMock(return_value=False)
    with patch("xonsh.ptk_shell.key_bindings.can_compile", mock):
        document = Document("for i in (1, 2, ")
        ctx.buffer.set_document(document)
        ctx.cr(ctx.buffer, ctx.cli)
        assert ctx.buffer.document.current_line == ""


def test_can_compile_and_executes(ctx):
    mock = MagicMock(return_value=True)
    with patch("xonsh.ptk_shell.key_bindings.can_compile", mock):
        document = Document("ls")
        ctx.buffer.set_document(document)
        ctx.cr(ctx.buffer, ctx.cli)
        assert ctx.accept.mock_calls is not None



import pytest


@pytest.fixture
def fish_completer(tmpdir, xession, load_xontrib, fake_process):
    """vox Alias function"""
    load_xontrib("fish_completer")
    xession.env.update(
        dict(
            XONSH_DATA_DIR=str(tmpdir),
            XONSH_SHOW_TRACEBACK=True,
        )
    )

    fake_process.register_subprocess(
        command=["fish", fake_process.any()],
        # completion for "git chec"
        stdout=b"""\
cherry-pick	Apply the change introduced by an existing commit
checkout	Checkout and switch to a branch""",
    )

    return fake_process


def test_fish_completer(fish_completer, check_completer):
    assert check_completer("git", prefix="chec") == {"checkout"}



import os
import pickle
import time
from unittest.mock import MagicMock

import pytest

from xonsh.commands_cache import (
    SHELL_PREDICTOR_PARSER,
    CommandsCache,
    predict_false,
    predict_shell,
    predict_true,
)
from xonsh.pytest.tools import skip_if_on_windows


def test_commands_cache_lazy(xession):
    cc = xession.commands_cache
    assert not cc.lazyin("xonsh")
    assert 0 == len(list(cc.lazyiter()))
    assert 0 == cc.lazylen()


def test_predict_threadable_unknown_command(xession):
    result = xession.commands_cache.predict_threadable(["command_should_not_found"])
    assert isinstance(result, bool)


@pytest.fixture
def commands_cache_tmp(xession, tmp_path, monkeypatch, patch_commands_cache_bins):
    xession.env["COMMANDS_CACHE_SAVE_INTERMEDIATE"] = True
    return patch_commands_cache_bins(["bin1", "bin2"])


def test_commands_cached_between_runs(commands_cache_tmp, tmp_path, tmpdir):
    # 1. no pickle file
    # 2. return empty result first and create a thread to populate result
    # 3. once the result is available then next call to cc.all_commands returns

    cc = commands_cache_tmp

    # wait for thread to end
    cnt = 0  # timeout waiting for thread
    while True:
        if cc.all_commands or cnt > 10:
            break
        cnt += 1
        time.sleep(0.1)
    assert [b.lower() for b in cc.all_commands.keys()] == ["bin1", "bin2"]

    files = tmp_path.glob("*.pickle")
    assert len(list(files)) == 1

    # cleanup dir
    for file in files:
        os.remove(file)


def test_commands_cache_uses_pickle_file(commands_cache_tmp, tmp_path, monkeypatch):
    cc = commands_cache_tmp
    update_cmds_cache = MagicMock()
    monkeypatch.setattr(cc, "_update_cmds_cache", update_cmds_cache)
    file = tmp_path / CommandsCache.CACHE_FILE
    bins = {
        "bin1": (
            "/some-path/bin1",
            None,
        ),
        "bin2": (
            "/some-path/bin2",
            None,
        ),
    }

    file.write_bytes(pickle.dumps(bins))
    assert str(cc.cache_file) == str(file)
    assert cc.all_commands == bins
    assert cc._loaded_pickled


TRUE_SHELL_ARGS = [
    ["-c", "yo"],
    ["-c=yo"],
    ["file"],
    ["-i", "-l", "file"],
    ["-i", "-c", "yo"],
    ["-i", "file"],
    ["-i", "-c", "yo", "file"],
]


@pytest.mark.parametrize("args", TRUE_SHELL_ARGS)
def test_predict_shell_parser(args):
    ns, unknown = SHELL_PREDICTOR_PARSER.parse_known_args(args)
    if ns.filename is not None:
        assert not ns.filename.startswith("-")


@pytest.mark.parametrize("args", TRUE_SHELL_ARGS)
def test_predict_shell_true(args):
    assert predict_shell(args)


FALSE_SHELL_ARGS = [[], ["-c"], ["-i"], ["-i", "-l"]]


@pytest.mark.parametrize("args", FALSE_SHELL_ARGS)
def test_predict_shell_false(args):
    assert not predict_shell(args)


PATTERN_BIN_USING_TTY_OR_NOT = [
    (
        False,
        {10: b"isnotatty"},
    ),
    (
        False,
        {12: b"isatty"},
    ),
    (
        False,
        {151: b"gpm"},
    ),
    (
        False,
        {10: b"isatty", 100: b"tcgetattr"},
    ),
    (
        False,
        {10: b"isatty", 100: b"tcsetattr"},
    ),
    (
        True,
        {10: b"isatty", 100: b"tcsetattr", 1000: b"tcgetattr"},
    ),
    (
        True,
        {1000: b"libncurses"},
    ),
    (
        True,
        {4094: b"libgpm"},
    ),
    (
        True,
        {2045: b"tcgetattr", 4095: b"tcgetattr", 6140: b"tcsetattr", 8190: b"isatty"},
    ),
]


@pytest.mark.parametrize("args", PATTERN_BIN_USING_TTY_OR_NOT)
@skip_if_on_windows
def test_commands_cache_predictor_default(args, xession, tmp_path):
    use_tty, patterns = args
    file = tmp_path / "testfile"
    where = list(patterns.keys())
    where.sort()

    with file.open("wb") as f:
        pos = 0
        for w in where:
            f.write(b"\x20" * (w - pos))
            f.write(patterns[w])
            pos = w + len(patterns[w])

        f.write(b"\x20" * (pos // 2))

    result = xession.commands_cache.default_predictor_readbin(
        "", str(file), timeout=1, failure=None
    )
    expected = predict_false if use_tty else predict_true
    assert result == expected


@skip_if_on_windows
def test_cd_is_only_functional_alias(xession):
    xession.aliases["cd"] = lambda args: os.chdir(args[0])
    xession.env["PATH"] = []
    assert xession.commands_cache.is_only_functional_alias("cd")


def test_non_exist_is_only_functional_alias(xession):
    assert not xession.commands_cache.is_only_functional_alias(
        "<not really a command name>"
    )


@skip_if_on_windows
def test_bash_is_only_functional_alias(xession):
    assert not xession.commands_cache.is_only_functional_alias("bash")


@skip_if_on_windows
def test_bash_and_is_alias_is_only_functional_alias(xession):
    xession.aliases["bash"] = lambda args: os.chdir(args[0])
    assert not xession.commands_cache.is_only_functional_alias("bash")



"""Tests the dummy history backend."""

import pytest

from xonsh.history.dummy import DummyHistory
from xonsh.history.main import construct_history


@pytest.mark.parametrize("backend", ["dummy", DummyHistory, DummyHistory()])
def test_construct_history_str(xession, backend):
    xession.env["XONSH_HISTORY_BACKEND"] = backend
    assert isinstance(construct_history(), DummyHistory)



import functools
from unittest.mock import Mock

import pytest

from xonsh.prompt import env as prompt_env
from xonsh.prompt.base import PromptFields, PromptFormatter


@pytest.fixture
def formatter(xession):
    return PromptFormatter()


@pytest.fixture
def live_fields(xession):
    return PromptFields(xession)


@pytest.mark.parametrize(
    "fields", [{"a_string": "cat", "none": (lambda: None), "f": (lambda: "wakka")}]
)
@pytest.mark.parametrize(
    "inp, exp",
    [
        ("my {a_string}", "my cat"),
        ("my {none}{a_string}", "my cat"),
        ("{f} jawaka", "wakka jawaka"),
    ],
)
def test_format_prompt(inp, exp, fields, formatter, xession):
    obs = formatter(template=inp, fields=fields)
    assert exp == obs


@pytest.mark.parametrize(
    "fields",
    [
        {
            "a_string": "cats",
            "a_number": 7,
            "empty": "",
            "current_job": (lambda: "sleep"),
            "none": (lambda: None),
        }
    ],
)
@pytest.mark.parametrize(
    "inp, exp",
    [
        ("{a_number:{0:^3}}cats", " 7 cats"),
        ("{current_job:{} | }xonsh", "sleep | xonsh"),
        ("{none:{} | }{a_string}{empty:!}", "cats!"),
        ("{none:{}}", ""),
        ("{{{a_string:{{{}}}}}}", "{{cats}}"),
        ("{{{none:{{{}}}}}}", "{}"),
    ],
)
def test_format_prompt_with_format_spec(inp, exp, fields, formatter):
    obs = formatter(template=inp, fields=fields)
    assert exp == obs


def test_format_prompt_with_broken_template(formatter):
    for p in ("{user", "{user}{hostname"):
        assert formatter(p) == p

    # '{{user' will be parsed to '{user'
    for p in ("{{user}", "{{user"):
        assert "user" in formatter(p)


@pytest.mark.parametrize("inp", ["{user", "{{user", "{{user}", "{user}{hostname"])
def test_format_prompt_with_broken_template_in_func(inp, formatter):
    # '{{user' will be parsed to '{user'
    assert "{user" in formatter(lambda: inp)


def test_format_prompt_with_invalid_func(formatter, xession):
    def p():
        foo = bar  # raises exception # noqa
        return "{user}"

    assert isinstance(formatter(p), str)


def test_format_prompt_with_func_that_raises(formatter, capsys, xession):
    template = "tt {zerodiv} tt"
    exp = "tt {BACKGROUND_RED}{ERROR:zerodiv}{RESET} tt"
    fields = {"zerodiv": lambda: 1 / 0}
    obs = formatter(template, fields)
    assert exp == obs
    out, err = capsys.readouterr()
    assert "prompt: error" in err


def test_format_prompt_with_no_env(formatter, xession, live_fields, env):
    xession.shell.prompt_formatter = formatter

    env.pop("VIRTUAL_ENV", None)  # For virtualenv
    env.pop("CONDA_DEFAULT_ENV", None)  # For conda/CircleCI

    assert formatter("{env_name}", fields=live_fields) == ""


@pytest.mark.parametrize("envname", ["env", "foo", "bar"])
def test_format_prompt_with_various_envs(formatter, xession, live_fields, envname):
    xession.shell.prompt_formatter = formatter

    xession.env["VIRTUAL_ENV"] = envname

    exp = live_fields["env_prefix"] + envname + live_fields["env_postfix"]
    assert formatter("{env_name}", fields=live_fields) == exp


@pytest.mark.parametrize("pre", ["(", "[[", "", "   "])
@pytest.mark.parametrize("post", [")", "]]", "", "   "])
def test_format_prompt_with_various_prepost(formatter, xession, live_fields, pre, post):
    xession.shell.prompt_formatter = formatter

    xession.env["VIRTUAL_ENV"] = "env"

    lf_copy = dict(live_fields)  # live_fields fixture is not idempotent!
    lf_copy.update({"env_prefix": pre, "env_postfix": post})
    exp = pre + "env" + post
    assert formatter("{env_name}", fields=lf_copy) == exp


def test_noenv_with_disable_set(formatter, xession, live_fields):
    xession.shell.prompt_formatter = formatter
    xession.env.update(dict(VIRTUAL_ENV="env", VIRTUAL_ENV_DISABLE_PROMPT=1))

    exp = ""
    assert formatter("{env_name}", fields=live_fields) == exp


class TestPromptFromVenvCfg:
    WANTED = "wanted"
    CONFIGS = [
        f"prompt = '{WANTED}'",
        f'prompt = "{WANTED}"',
        f'\t prompt =  "{WANTED}"   ',
        f"prompt \t=    {WANTED}   ",
        "nothing = here",
    ]
    CONFIGS.extend([f"other = fluff\n{t}\nmore = fluff" for t in CONFIGS])

    @pytest.mark.parametrize("text", CONFIGS)
    def test_determine_env_name_from_cfg(self, monkeypatch, tmp_path, text):
        monkeypatch.setattr(prompt_env, "_surround_env_name", lambda x: x)
        (tmp_path / "pyvenv.cfg").write_text(text)
        wanted = self.WANTED if self.WANTED in text else tmp_path.name
        assert prompt_env._determine_env_name(tmp_path) == wanted


class TestEnvNamePrompt:
    def test_no_prompt(self, formatter, live_fields):
        assert formatter("{env_name}", fields=live_fields) == ""

    def test_search_order(self, monkeypatch, tmp_path, formatter, xession, live_fields):
        xession.shell.prompt_formatter = formatter

        first = "first"
        second = "second"
        third = "third"
        fourth = "fourth"

        pyvenv_cfg = tmp_path / third / "pyvenv.cfg"
        pyvenv_cfg.parent.mkdir()
        pyvenv_cfg.write_text(f"prompt={second}")

        fmt = functools.partial(formatter, "{env_name}", fields=live_fields)
        xession.env.update(
            dict(
                VIRTUAL_ENV_PROMPT=first,
                VIRTUAL_ENV=str(pyvenv_cfg.parent),
                CONDA_DEFAULT_ENV=fourth,
            )
        )

        xession.env["VIRTUAL_ENV_DISABLE_PROMPT"] = 0
        assert fmt() == first

        live_fields.reset()
        xession.env["VIRTUAL_ENV_DISABLE_PROMPT"] = 1
        assert fmt() == ""

        live_fields.reset()
        del xession.env["VIRTUAL_ENV_PROMPT"]
        xession.env["VIRTUAL_ENV_DISABLE_PROMPT"] = 0
        assert fmt() == f"({second}) "

        live_fields.reset()
        xession.env["VIRTUAL_ENV_DISABLE_PROMPT"] = 1
        assert fmt() == ""

        live_fields.reset()
        xession.env["VIRTUAL_ENV_DISABLE_PROMPT"] = 0
        pyvenv_cfg.unlink()
        # In the interest of speed the calls are cached, but if the user
        # fiddles with environments this will bite them. I will not do anythin
        prompt_env._determine_env_name.cache_clear()
        assert fmt() == f"({third}) "

        live_fields.reset()
        xession.env["VIRTUAL_ENV_DISABLE_PROMPT"] = 1
        assert fmt() == ""

        live_fields.reset()
        xession.env["VIRTUAL_ENV_DISABLE_PROMPT"] = 0
        del xession.env["VIRTUAL_ENV"]
        assert fmt() == f"({fourth}) "

        live_fields.reset()
        xession.env["VIRTUAL_ENV_DISABLE_PROMPT"] = 1
        assert fmt() == ""

        live_fields.reset()
        xession.env["VIRTUAL_ENV_DISABLE_PROMPT"] = 0
        del xession.env["CONDA_DEFAULT_ENV"]
        assert fmt() == ""

    @pytest.mark.xfail(reason="caching introduces stale values")
    def test_env_name_updates_on_filesystem_change(self, tmp_path):
        """Due to cache, user might get stale value.

        if user fiddles with env folder or the config, they might get a stale
        value from the cache.
        """
        cfg = tmp_path / "pyvenv.cfg"
        cfg.write_text("prompt=fromfile")
        assert prompt_env._determine_env_name(cfg.parent) == "fromfile"
        cfg.unlink()
        assert prompt_env._determine_env_name(cfg.parent) == cfg.parent.name


@pytest.mark.parametrize("disable", [0, 1])
def test_custom_env_overrides_default(formatter, xession, live_fields, disable):
    xession.shell.prompt_formatter = formatter

    prompt = "!venv active! "

    xession.env.update(
        dict(
            VIRTUAL_ENV="env",
            VIRTUAL_ENV_PROMPT=prompt,
            VIRTUAL_ENV_DISABLE_PROMPT=disable,
        )
    )

    exp = "" if disable else prompt
    assert formatter("{env_name}", fields=live_fields) == exp


def test_promptformatter_cache(formatter):
    spam = Mock()
    template = "{spam} and {spam}"
    fields = {"spam": spam}

    formatter(template, fields)

    assert spam.call_count == 1


def test_promptformatter_clears_cache(formatter):
    spam = Mock()
    template = "{spam} and {spam}"
    fields = {"spam": spam}

    formatter(template, fields)
    formatter(template, fields)

    assert spam.call_count == 2



"""A (tab-)completer for xonsh."""
import collections.abc as cabc
import sys
import typing as tp

from xonsh.built_ins import XSH
from xonsh.completers.tools import (
    Completion,
    RichCompletion,
    apply_lprefix,
    get_filter_function,
    is_contextual_completer,
    is_exclusive_completer,
)
from xonsh.parsers.completion_context import CompletionContext, CompletionContextParser
from xonsh.tools import print_exception


class Completer:
    """This provides a list of optional completions for the xonsh shell."""

    def __init__(self):
        self.context_parser = CompletionContextParser()

    def parse(
        self, text: str, cursor_index: "None|int" = None, ctx=None
    ) -> "CompletionContext":
        """Parse the given text

        Parameters
        ----------
        text
            multi-line text
        cursor_index
            position of the cursor. If not given, then it is considered to be at the end.
        ctx
            Execution context
        """
        cursor_index = len(text) if cursor_index is None else cursor_index
        return self.context_parser.parse(text, cursor_index, ctx)

    def complete_line(self, text: str):
        """Handy wrapper to build command-completion-context when cursor is at the end.

        Notes
        -----
        suffix is not supported; text after last space is parsed as prefix.
        """
        ctx = self.parse(text)
        cmd_ctx = ctx.command
        if not cmd_ctx:
            raise RuntimeError("Only Command context is empty")
        prefix = cmd_ctx.prefix

        line = text
        begidx = text.rfind(prefix)
        endidx = begidx + len(prefix)

        return self.complete(
            prefix,
            line,
            begidx,
            endidx,
            cursor_index=len(line),
            multiline_text=line,
            completion_context=ctx,
        )

    def complete(
        self,
        prefix,
        line,
        begidx,
        endidx,
        ctx=None,
        multiline_text=None,
        cursor_index=None,
        completion_context=None,
    ):
        """Complete the string, given a possible execution context.

        Parameters
        ----------
        prefix : str
            The string to match
        line : str
            The line that prefix appears on.
        begidx : int
            The index in line that prefix starts on.
        endidx : int
            The index in line that prefix ends on.
        ctx : dict, optional
            Names in the current execution context.
        multiline_text : str
            The complete multiline text. Needed to get completion context.
        cursor_index : int
            The current cursor's index in the multiline text.
            May be ``len(multiline_text)`` for cursor at the end.
            Needed to get completion context.

        Returns
        -------
        rtn : list of str
            Possible completions of prefix, sorted alphabetically.
        lprefix : int
            Length of the prefix to be replaced in the completion.
        """

        if (
            (multiline_text is not None)
            and (cursor_index is not None)
            and (completion_context is None)
        ):
            completion_context: tp.Optional[CompletionContext] = self.parse(
                multiline_text,
                cursor_index,
                ctx,
            )

        ctx = ctx or {}
        return self.complete_from_context(
            completion_context,
            (prefix, line, begidx, endidx, ctx),
        )

    @staticmethod
    def _format_completion(
        completion,
        completion_context,
        completing_contextual_command: bool,
        lprefix: int,
        custom_lprefix: bool,
    ) -> tp.Tuple[Completion, int]:
        if (
            completing_contextual_command
            and completion_context.command.is_after_closing_quote
        ):
            """
            The cursor is appending to a closed string literal, i.e. cursor at the end of ``ls "/usr/"``.
            1. The closing quote will be appended to all completions.
                I.e the completion ``/usr/bin`` will turn into ``/usr/bin"``
                To prevent this behavior, a completer can return a ``RichCompletion`` with ``append_closing_quote=False``.
            2. If not specified, lprefix will cover the closing prefix.
                I.e for ``ls "/usr/"``, the default lprefix will be 6 to include the closing quote.
                To prevent this behavior, a completer can return a different lprefix or specify it inside ``RichCompletion``.
            """
            closing_quote = completion_context.command.closing_quote
            if not custom_lprefix:
                lprefix += len(closing_quote)
            if closing_quote:
                if isinstance(completion, RichCompletion):
                    if completion.append_closing_quote:
                        completion = completion.replace(
                            value=completion.value + closing_quote
                        )
                else:
                    completion = completion + closing_quote

        completion = list(apply_lprefix([completion], lprefix))[0]

        if (
            isinstance(completion, RichCompletion)
            and completion.append_space
            and not completion.value.endswith(" ")
        ):
            # append spaces AFTER appending closing quote
            completion = completion.replace(value=completion.value + " ")

        return completion, lprefix

    @staticmethod
    def generate_completions(
        completion_context, old_completer_args, trace: bool
    ) -> tp.Iterator[tp.Tuple[Completion, int]]:
        filter_func = get_filter_function()

        for name, func in XSH.completers.items():
            try:
                if is_contextual_completer(func):
                    if completion_context is None:
                        continue
                    out = func(completion_context)
                else:
                    if old_completer_args is None:
                        continue
                    out = func(*old_completer_args)
            except StopIteration:
                # completer requested to stop collecting completions
                break
            except Exception as e:
                name = func.__name__ if hasattr(func, "__name__") else str(func)
                print_exception(
                    f"Completer {name} raises exception when gets "
                    f"old_args={old_completer_args[:-1]} / completion_context={completion_context!r}:\n"
                    f"{type(e)} - {e}"
                )
                continue

            completing_contextual_command = (
                is_contextual_completer(func)
                and completion_context is not None
                and completion_context.command is not None
            )

            # -- set comp-defaults --

            # the default is that the completer function filters out as necessary
            # we can change that once fuzzy/substring matches are added
            is_filtered = True
            custom_lprefix = False
            prefix = ""
            if completing_contextual_command:
                prefix = completion_context.command.prefix
            elif old_completer_args is not None:
                prefix = old_completer_args[0]
            lprefix = len(prefix)

            if isinstance(out, cabc.Sequence):
                # update comp-defaults from
                res, lprefix_filtered = out
                if isinstance(lprefix_filtered, bool):
                    is_filtered = lprefix_filtered
                else:
                    lprefix = lprefix_filtered
                    custom_lprefix = True
            else:
                res = out

            if res is None:
                continue

            items = []
            for comp in res:
                if (not is_filtered) and (not filter_func(comp, prefix)):
                    continue
                comp = Completer._format_completion(
                    comp,
                    completion_context,
                    completing_contextual_command,
                    lprefix or 0,
                    custom_lprefix,
                )
                items.append(comp)
                yield comp

            if not items:  # empty completion
                continue

            if trace:
                print(
                    f"TRACE COMPLETIONS: Got {len(items)} results"
                    f" from {'' if is_exclusive_completer(func) else 'non-'}exclusive completer '{name}':"
                )
                sys.displayhook(items)

            if is_exclusive_completer(func):
                # we got completions for an exclusive completer
                break

    def complete_from_context(self, completion_context, old_completer_args=None):
        trace = XSH.env.get("XONSH_TRACE_COMPLETIONS")
        if trace:
            print("\nTRACE COMPLETIONS: Getting completions with context:")
            sys.displayhook(completion_context)
        lprefix = 0

        # using dict to keep order py3.6+
        completions = {}

        query_limit = XSH.env.get("COMPLETION_QUERY_LIMIT")

        for comp in self.generate_completions(
            completion_context,
            old_completer_args,
            trace,
        ):
            completion, lprefix = comp
            completions[completion] = None
            if query_limit and len(completions) >= query_limit:
                if trace:
                    print(
                        "TRACE COMPLETIONS: Stopped after $COMPLETION_QUERY_LIMIT reached."
                    )
                break

        def sortkey(s):
            # todo: should sort with prefix > substring > fuzzy
            return s.lstrip(''''"''').lower()

        # the last completer's lprefix is returned. other lprefix values are inside the RichCompletions.
        return tuple(sorted(completions, key=sortkey)), lprefix



import itertools
import typing as tp
from unittest import mock

import pytest

import xonsh.parsers.completion_context as ctx
from xonsh.parsers.completion_context import (
    CommandArg,
    CommandContext,
    CompletionContextParser,
    PythonContext,
)
from xonsh.pytest.tools import ON_WINDOWS

DEBUG = False
MISSING = object()
X = "\x00"  # cursor marker
PARSER: tp.Optional[CompletionContextParser] = None


@pytest.fixture(scope="module", autouse=True)
def parser():
    global PARSER
    PARSER = CompletionContextParser(debug=DEBUG)
    patcher = None
    if ON_WINDOWS:
        # on-windows has an option for interactive sessions. Overriding the lazyObject
        patcher = mock.patch.object(
            ctx,
            "LINE_CONT_REPLACEMENT_DIFF",
            ("\\\n", "", -2),
        )
        patcher.start()
    yield
    PARSER = None
    if ON_WINDOWS and patcher:
        patcher.stop()


def parse(command, inner_index):
    return PARSER.parse(command, inner_index)


def assert_match(
    commandline, command_context=MISSING, python_context=MISSING, is_main_command=False
):
    if X in commandline:
        index = commandline.index(X)
        commandline = commandline.replace(X, "")
    else:
        index = len(commandline)
    context = parse(commandline, index)
    if context is None:
        raise SyntaxError(
            "Failed to parse the commandline - set DEBUG = True in this file to see the error"
        )
    if is_main_command and python_context is MISSING:
        python_context = PythonContext(commandline, index)
    if command_context is not MISSING:
        assert context.command == command_context
    if python_context is not MISSING:
        assert context.python == python_context


COMMAND_EXAMPLES = (
    (f"comm{X}", CommandContext(args=(), arg_index=0, prefix="comm")),
    (f" comm{X}", CommandContext(args=(), arg_index=0, prefix="comm")),
    (f"comm{X}and", CommandContext(args=(), arg_index=0, prefix="comm", suffix="and")),
    (f"command {X}", CommandContext(args=(CommandArg("command"),), arg_index=1)),
    (f"{X} command", CommandContext(args=(CommandArg("command"),), arg_index=0)),
    (f" command {X}", CommandContext(args=(CommandArg("command"),), arg_index=1)),
    (
        f"command --{X}",
        CommandContext(args=(CommandArg("command"),), arg_index=1, prefix="--"),
    ),
    (
        f"command a {X}",
        CommandContext(args=(CommandArg("command"), CommandArg("a")), arg_index=2),
    ),
    (
        f"command a b{X}",
        CommandContext(
            args=(CommandArg("command"), CommandArg("a")), arg_index=2, prefix="b"
        ),
    ),
    (
        f"command a   b{X}",
        CommandContext(
            args=(CommandArg("command"), CommandArg("a")), arg_index=2, prefix="b"
        ),
    ),
    (
        f"command {X} a",
        CommandContext(args=(CommandArg("command"), CommandArg("a")), arg_index=1),
    ),
    (
        f"command a {X} b",
        CommandContext(
            args=(CommandArg("command"), CommandArg("a"), CommandArg("b")), arg_index=2
        ),
    ),
    (
        f"command -{X} b",
        CommandContext(
            args=(CommandArg("command"), CommandArg("b")), arg_index=1, prefix="-"
        ),
    ),
    (
        f"command a {X}b",
        CommandContext(
            args=(CommandArg("command"), CommandArg("a")), arg_index=2, suffix="b"
        ),
    ),
    (
        f"command a{X}b",
        CommandContext(
            args=(CommandArg("command"),), arg_index=1, prefix="a", suffix="b"
        ),
    ),
    (
        f"'comm and' a{X}b",
        CommandContext(
            args=(CommandArg("comm and", opening_quote="'", closing_quote="'"),),
            arg_index=1,
            prefix="a",
            suffix="b",
        ),
    ),
)

EMPTY_COMMAND_EXAMPLES = (
    (f"{X}", CommandContext((), 0)),
    (f" {X}", CommandContext((), 0)),
    (f"{X} ", CommandContext((), 0)),
    (f" {X} ", CommandContext((), 0)),
)

STRING_ARGS_EXAMPLES = (
    (
        f"'comm an{X}d'",
        CommandContext(
            args=(),
            arg_index=0,
            prefix="comm an",
            suffix="d",
            opening_quote="'",
            closing_quote="'",
        ),
    ),
    (
        f"'comm and{X}'",
        CommandContext(
            args=(),
            arg_index=0,
            prefix="comm and",
            suffix="",
            opening_quote="'",
            closing_quote="'",
        ),
    ),
    (
        f"'comm {X}'",
        CommandContext(
            args=(),
            arg_index=0,
            prefix="comm ",
            suffix="",
            opening_quote="'",
            closing_quote="'",
        ),
    ),
    (
        f'"comm an{X}d"',
        CommandContext(
            args=(),
            arg_index=0,
            prefix="comm an",
            suffix="d",
            opening_quote='"',
            closing_quote='"',
        ),
    ),
    (
        f"'''comm an{X}d'''",
        CommandContext(
            args=(),
            arg_index=0,
            prefix="comm an",
            suffix="d",
            opening_quote="'''",
            closing_quote="'''",
        ),
    ),
    (
        f"fr'comm an{X}d'",
        CommandContext(
            args=(),
            arg_index=0,
            prefix="comm an",
            suffix="d",
            opening_quote="fr'",
            closing_quote="'",
        ),
    ),
    (
        f"'()+{X}'",
        CommandContext(
            args=(), arg_index=0, prefix="()+", opening_quote="'", closing_quote="'"
        ),
    ),
    (
        f"'comm and'{X}",
        CommandContext(
            args=(),
            arg_index=0,
            prefix="comm and",
            opening_quote="'",
            closing_quote="'",
            is_after_closing_quote=True,
        ),
    ),
    (
        f"'''comm and'''{X}",
        CommandContext(
            args=(),
            arg_index=0,
            prefix="comm and",
            opening_quote="'''",
            closing_quote="'''",
            is_after_closing_quote=True,
        ),
    ),
)

COMMAND_EXAMPLES += STRING_ARGS_EXAMPLES
COMMAND_EXAMPLES += EMPTY_COMMAND_EXAMPLES


@pytest.mark.parametrize("commandline, context", COMMAND_EXAMPLES)
def test_command(commandline, context):
    assert_match(commandline, context, is_main_command=True)


@pytest.mark.parametrize(
    "commandline, context",
    tuple(
        (commandline, context)
        for commandline, context in STRING_ARGS_EXAMPLES
        if commandline.endswith("'") or commandline.endswith('"')
    ),
)
def test_partial_string_arg(commandline, context):
    partial_commandline = commandline.rstrip("\"'")
    partial_context = context._replace(closing_quote="")
    assert_match(partial_commandline, partial_context, is_main_command=True)


CONT = "\\" "\n"


@pytest.mark.parametrize(
    "commandline, context",
    (
        # line continuations:
        (
            f"echo {CONT}a {X}",
            CommandContext(args=(CommandArg("echo"), CommandArg("a")), arg_index=2),
        ),
        (
            f"echo {CONT}{X}a {CONT} b",
            CommandContext(
                args=(CommandArg("echo"), CommandArg("b")), arg_index=1, suffix="a"
            ),
        ),
        (
            f"echo a{CONT}{X}b",
            CommandContext(
                args=(CommandArg("echo"),), arg_index=1, prefix="a", suffix="b"
            ),
        ),
        (
            f"echo a{X}{CONT}b",
            CommandContext(
                args=(CommandArg("echo"),), arg_index=1, prefix="a", suffix="b"
            ),
        ),
        (
            f"echo ${CONT}(a) {CONT} {X}b",
            CommandContext(
                args=(CommandArg("echo"), CommandArg("$(a)")), arg_index=2, suffix="b"
            ),
        ),
        # line continuations in strings:
        (
            f"echo 'a{CONT}{X}b'",
            CommandContext(
                args=(CommandArg("echo"),),
                arg_index=1,
                prefix="a",
                suffix="b",
                opening_quote="'",
                closing_quote="'",
            ),
        ),
        (
            f"echo '''a{CONT}{X}b'''",
            CommandContext(
                args=(CommandArg("echo"),),
                arg_index=1,
                prefix="a",
                suffix="b",
                opening_quote="'''",
                closing_quote="'''",
            ),
        ),
        (
            f"echo 'a{CONT}{X}b",
            CommandContext(
                args=(CommandArg("echo"),),
                arg_index=1,
                prefix="a",
                suffix="b",
                opening_quote="'",
            ),
        ),
        (
            f"echo '''a{CONT}{X}b",
            CommandContext(
                args=(CommandArg("echo"),),
                arg_index=1,
                prefix="a",
                suffix="b",
                opening_quote="'''",
            ),
        ),
        (
            f"echo ''{CONT}'a{X}b",
            CommandContext(
                args=(CommandArg("echo"),),
                arg_index=1,
                prefix="a",
                suffix="b",
                opening_quote="'''",
            ),
        ),
        (
            f"echo '''a{CONT}{X} b",
            CommandContext(
                args=(CommandArg("echo"),),
                arg_index=1,
                prefix="a",
                suffix=" b",
                opening_quote="'''",
            ),
        ),
        # triple-quoted strings:
        (
            f"echo '''a\nb{X}\nc'''",
            CommandContext(
                args=(CommandArg("echo"),),
                arg_index=1,
                prefix="a\nb",
                suffix="\nc",
                opening_quote="'''",
                closing_quote="'''",
            ),
        ),
        (
            f"echo '''a\n b{X} \n  c'''",
            CommandContext(
                args=(CommandArg("echo"),),
                arg_index=1,
                prefix="a\n b",
                suffix=" \n  c",
                opening_quote="'''",
                closing_quote="'''",
            ),
        ),
        # partial triple-quoted strings:
        (
            f"echo '''a\nb{X}\nc",
            CommandContext(
                args=(CommandArg("echo"),),
                arg_index=1,
                prefix="a\nb",
                suffix="\nc",
                opening_quote="'''",
            ),
        ),
        (
            f"echo '''a\n b{X} \n  c",
            CommandContext(
                args=(CommandArg("echo"),),
                arg_index=1,
                prefix="a\n b",
                suffix=" \n  c",
                opening_quote="'''",
            ),
        ),
    ),
)
def test_multiline_command(commandline, context):
    assert_match(commandline, context, is_main_command=True)


NESTING_EXAMPLES = (
    # nesting, prefix
    (f"echo $({X})", "$("),
    (f"echo @$({X})", "@$("),
    (f"echo $(echo $({X}))", "$("),
    (f"echo @(x + $({X}))", "$("),
    (f"!({X})", "!("),
    (f"$[{X}]", "$["),
    (f"![{X}]", "!["),
)

NESTED_SIMPLE_CMD_EXAMPLES = [
    (
        nesting,
        f"simple {X}",
        CommandContext(
            args=(CommandArg("simple"),), arg_index=1, subcmd_opening=prefix
        ),
    )
    for nesting, prefix in NESTING_EXAMPLES[1:]
]


@pytest.mark.parametrize(
    "nesting, commandline, context",
    list(
        itertools.chain(
            (
                # complex subcommand in a simple nested expression
                (
                    NESTING_EXAMPLES[0][0],
                    commandline,
                    context._replace(subcmd_opening=NESTING_EXAMPLES[0][1]),
                )
                for commandline, context in COMMAND_EXAMPLES
            ),
            NESTED_SIMPLE_CMD_EXAMPLES,
        )
    ),
)
def test_nested_command(commandline, context, nesting):
    nested_commandline = nesting.replace(X, commandline)
    assert_match(nested_commandline, command_context=context, python_context=None)


NESTING_MALFORMATIONS = (
    lambda s: s[:-1],  # remove the last closing brace ')' / ']'
    lambda s: s + s[-1],  # add an extra closing brace ')' / ']'
    lambda s: s[-1] + s,
    lambda s: s + "$(",
    lambda s: "$(" + s,
)


@pytest.mark.parametrize("nesting, commandline, context", NESTED_SIMPLE_CMD_EXAMPLES)
@pytest.mark.parametrize("malformation", NESTING_MALFORMATIONS)
def test_malformed_subcmd(nesting, commandline, context, malformation):
    nested_commandline = nesting.replace(X, commandline)
    nested_commandline = malformation(nested_commandline)
    assert_match(nested_commandline, command_context=context, python_context=None)


MALFORMED_SUBCOMMANDS_NESTINGS = (
    # nesting, subcmd_opening
    (f"echo $(a $({X}", "$("),
    (f"echo $(a $(b; {X}", ""),
    (f"$(echo $(a $({X}", "$("),
    (f"echo $[a $({X}]", "$("),
    (f"echo $(a $[{X})", "$["),
    (f"echo @(x = $({X}", "$("),
    (f"echo @(a; x = $({X}", "$("),
    (f"echo @(x = $(a; {X}", ""),
)


@pytest.mark.parametrize("nesting, subcmd_opening", MALFORMED_SUBCOMMANDS_NESTINGS)
@pytest.mark.parametrize("commandline, context", COMMAND_EXAMPLES[:5])
def test_multiple_malformed_subcmds(nesting, subcmd_opening, commandline, context):
    nested_commandline = nesting.replace(X, commandline)
    nested_context = context._replace(subcmd_opening=subcmd_opening)
    assert_match(nested_commandline, nested_context, python_context=None)


def test_other_subcommand_arg():
    command = "echo $(pwd) "
    assert_match(
        command,
        CommandContext((CommandArg("echo"), CommandArg("$(pwd)")), arg_index=2),
        is_main_command=True,
    )


def test_combined_subcommand_arg():
    command = f"echo file=$(pwd{X})/x"

    # index inside the subproc
    assert_match(
        command,
        CommandContext((), arg_index=0, prefix="pwd", subcmd_opening="$("),
        python_context=None,
    )

    # index at the end of the command
    assert_match(
        command.replace(X, ""),
        CommandContext((CommandArg("echo"),), arg_index=1, prefix="file=$(pwd)/x"),
        is_main_command=True,
    )


SUBCMD_BORDER_EXAMPLES = (
    (f"{X}$(echo)", CommandContext((), 0, suffix="$(echo)")),
    (f"${X}(echo)", CommandContext((), 0, prefix="$", suffix="(echo)")),
    (f"$(echo){X}", CommandContext((), 0, prefix="$(echo)")),
    (f"${X}( echo)", CommandContext((), 0, prefix="$", suffix="( echo)")),
    (f"$(echo ){X}", CommandContext((), 0, prefix="$(echo )")),
)


@pytest.mark.parametrize("commandline, context", SUBCMD_BORDER_EXAMPLES)
def test_cursor_in_subcmd_borders(commandline, context):
    assert_match(commandline, context, is_main_command=True)


MULTIPLE_COMMAND_KEYWORDS = (
    "; ",
    "\n",
    " and ",
    "&& ",
    " or ",
    "|| ",
    "| ",
)

MULTIPLE_CMD_SIMPLE_EXAMPLES = [
    (
        keyword,
        ("echo hi", f"simple {X}"),
        CommandContext(args=(CommandArg("simple"),), arg_index=1),
    )
    for keyword in MULTIPLE_COMMAND_KEYWORDS
]

EXTENSIVE_COMMAND_PAIRS = tuple(
    itertools.chain(
        zip(COMMAND_EXAMPLES, COMMAND_EXAMPLES[::-1]),
        zip(COMMAND_EXAMPLES, EMPTY_COMMAND_EXAMPLES),
        zip(EMPTY_COMMAND_EXAMPLES, COMMAND_EXAMPLES),
        zip(EMPTY_COMMAND_EXAMPLES, EMPTY_COMMAND_EXAMPLES),
    )
)

MULTIPLE_COMMAND_EXTENSIVE_EXAMPLES = tuple(
    itertools.chain(
        (
            # cursor in first command
            ((first, second.replace(X, "")), first_context)
            for (first, first_context), (
                second,
                second_context,
            ) in EXTENSIVE_COMMAND_PAIRS
        ),
        (
            # cursor in second command
            ((first.replace(X, ""), second), second_context)
            for (first, first_context), (
                second,
                second_context,
            ) in EXTENSIVE_COMMAND_PAIRS
        ),
        (
            # cursor in middle command
            ((first.replace(X, ""), second, third.replace(X, "")), second_context)
            for (first, _1), (second, second_context), (third, _3) in zip(
                COMMAND_EXAMPLES[:3], COMMAND_EXAMPLES[3:6], COMMAND_EXAMPLES[6:9]
            )
        ),
        (
            # cursor in third command
            ((first.replace(X, ""), second.replace(X, ""), third), third_context)
            for (first, _1), (second, _2), (third, third_context) in zip(
                COMMAND_EXAMPLES[:3], COMMAND_EXAMPLES[3:6], COMMAND_EXAMPLES[6:9]
            )
        ),
    )
)


@pytest.mark.parametrize(
    "keyword, commands, context",
    tuple(
        itertools.chain(
            (
                (MULTIPLE_COMMAND_KEYWORDS[0], commands, context)
                for commands, context in MULTIPLE_COMMAND_EXTENSIVE_EXAMPLES
            ),
            MULTIPLE_CMD_SIMPLE_EXAMPLES,
        )
    ),
)
def test_multiple_commands(keyword, commands, context):
    joined_command = keyword.join(commands)

    cursor_command = next(command for command in commands if X in command)
    if cursor_command is commands[0]:
        relative_index = cursor_command.index(X)
    else:
        absolute_index = joined_command.index(X)
        relative_index = (
            absolute_index
            - joined_command.rindex(keyword, 0, absolute_index)
            - len(keyword)
        )
        if keyword.endswith(" "):
            # the last space is part of the command
            relative_index += 1
            cursor_command = " " + cursor_command

    assert_match(joined_command, context, is_main_command=True)


@pytest.mark.parametrize(
    "commandline",
    (
        f"{X};",
        f"; {X}",
        f"{X};;",
        f"; {X};",
        f";; {X}",
        f";;; {X}",
    ),
)
def test_multiple_empty_commands(commandline):
    assert_match(commandline, CommandContext((), 0), is_main_command=True)


@pytest.mark.parametrize(
    "nesting, keyword, commands, context",
    tuple(
        (
            nesting,
            keyword,
            commands,
            context,
        )  # no subcmd_opening in nested multi-commands
        for nesting, prefix in NESTING_EXAMPLES
        for keyword, commands, context in MULTIPLE_CMD_SIMPLE_EXAMPLES
        if keyword != "\n"  # the lexer ignores newlines inside subcommands
    ),
)
def test_nested_multiple_commands(nesting, keyword, commands, context):
    joined_command = keyword.join(commands)
    nested_joined = nesting.replace(X, joined_command)
    assert_match(nested_joined, context, python_context=None)


def test_multiple_nested_commands():
    assert_match(
        f"echo hi; echo $(ls{X})",
        CommandContext((), 0, prefix="ls", subcmd_opening="$("),
        python_context=None,
    )


@pytest.mark.parametrize(
    "commandline, context",
    tuple(
        (commandline, context)
        for commandline, context in STRING_ARGS_EXAMPLES
        if commandline.endswith("'") or commandline.endswith('"')
    ),
)
def test_multiple_partial_string_arg(commandline, context):
    partial_commandline = commandline.rstrip("\"'")
    partial_context = context._replace(closing_quote="")
    assert_match("echo;" + partial_commandline, partial_context)
    assert_match("echo $[a ;" + partial_commandline, partial_context)


@pytest.mark.parametrize(
    "nesting, keyword, commands, context",
    tuple(
        (nesting, keyword, commands, context)
        for nesting, prefix in NESTING_EXAMPLES[:1]
        for keyword, commands, context in MULTIPLE_CMD_SIMPLE_EXAMPLES[:1]
    ),
)
@pytest.mark.parametrize("malformation", NESTING_MALFORMATIONS)
def test_malformed_subcmd_1(malformation, nesting, keyword, commands, context):
    joined_command = keyword.join(commands)
    nested_joined = nesting.replace(X, joined_command)
    malformed_commandline = malformation(nested_joined)
    assert_match(malformed_commandline, context, python_context=None)


MULTIPLE_COMMAND_BORDER_EXAMPLES = tuple(
    itertools.chain(
        itertools.chain(
            *(
                (
                    (
                        f"ls{ws1}{X}{kwd}{ws2}echo",
                        CommandContext((CommandArg("ls"),), 1)
                        if ws1
                        else CommandContext((), 0, prefix="ls"),
                    ),
                )
                for ws1, ws2, kwd in itertools.product(
                    ("", " "), ("", " "), ("&&", ";")
                )
            )
        ),
        # all keywords are treated as a normal arg if the cursor is at the edge
        (
            (
                f"ls {X}and echo",
                CommandContext((CommandArg("ls"), CommandArg("echo")), 1, suffix="and"),
            ),
            (
                f"ls and{X} echo",
                CommandContext((CommandArg("ls"), CommandArg("echo")), 1, prefix="and"),
            ),
            (
                f"ls ||{X} echo",
                CommandContext((CommandArg("ls"), CommandArg("echo")), 1, prefix="||"),
            ),
        ),
        # if the cursor is inside the keyword, it's treated as a normal arg
        (
            (
                f"ls a{X}nd echo",
                CommandContext(
                    (CommandArg("ls"), CommandArg("echo")), 1, prefix="a", suffix="nd"
                ),
            ),
            (
                f"ls &{X}& echo",
                CommandContext(
                    (CommandArg("ls"), CommandArg("echo")), 1, prefix="&", suffix="&"
                ),
            ),
        ),
    )
)


@pytest.mark.parametrize(
    "commandline, context",
    tuple(
        itertools.chain(
            MULTIPLE_COMMAND_BORDER_EXAMPLES,
            (
                # ensure these rules work with more than one command
                (f"cat | {commandline}", context)
                for commandline, context in MULTIPLE_COMMAND_BORDER_EXAMPLES
            ),
        )
    ),
)
def test_cursor_in_multiple_keyword_borders(commandline, context):
    assert_match(commandline, context)


PYTHON_EXAMPLES = (
    # commandline, context
    (f"x = {X}", PythonContext("x = ", 4)),
    (f"a {X}= x; b = y", PythonContext("a = x; b = y", 2)),
    (f"a {X}= x\nb = $(ls)", PythonContext("a = x\nb = $(ls)", 2)),
)

PYTHON_NESTING_EXAMPLES = (
    # nesting, prefix
    f"echo @({X})",
    f"echo $(echo @({X}))",
    f"echo @(x + @({X}))",  # invalid syntax, but can still be in a partial command
)


@pytest.mark.parametrize(
    "nesting, commandline, context",
    list(
        itertools.chain(
            (
                # complex subcommand in a simple nested expression
                (nesting, commandline, context._replace(is_sub_expression=True))
                for nesting in PYTHON_NESTING_EXAMPLES[:1]
                for commandline, context in PYTHON_EXAMPLES
            ),
            (
                # simple subcommand in a complex nested expression
                (nesting, commandline, context._replace(is_sub_expression=True))
                for nesting in PYTHON_NESTING_EXAMPLES
                for commandline, context in PYTHON_EXAMPLES[:1]
            ),
        )
    ),
)
def test_nested_python(commandline, context, nesting):
    nested_commandline = nesting.replace(X, commandline)
    assert_match(nested_commandline, command_context=None, python_context=context)


@pytest.mark.parametrize(
    "commandline, context",
    [
        (
            commandline.replace("$", "@"),
            context._replace(
                prefix=context.prefix.replace("$", "@"),
                suffix=context.suffix.replace("$", "@"),
            ),
        )
        for commandline, context in SUBCMD_BORDER_EXAMPLES
    ],
)
def test_cursor_in_sub_python_borders(commandline, context):
    assert_match(commandline, context, is_main_command=True)


@pytest.mark.parametrize(
    "code",
    (
        f"""
x = 3
x.{X}""",
        f"""
x = 3;
y = 4;
x.{X}""",
        f"""
def func({X}):
    return 100
    """,
        f"""
class A:
    def a():
        return "a{X}"
    pass
exit()
    """,
    ),
)
def test_multiline_python(code):
    assert_match(code, is_main_command=True)



import builtins
import os.path
from contextlib import contextmanager
from unittest.mock import MagicMock

import pytest

from xonsh.aliases import make_default_aliases, source_alias


@pytest.fixture
def mockopen(xession, monkeypatch):
    @contextmanager
    def mocked_open(fpath, *args, **kwargs):
        yield MagicMock(read=lambda: fpath)

    monkeypatch.setattr(builtins, "open", mocked_open)


@pytest.fixture
def mocked_execx_checker(xession, monkeypatch):
    checker = []

    def mocked_execx(src, *args, **kwargs):
        checker.append(src.strip())

    monkeypatch.setattr(xession.builtins, "execx", mocked_execx)
    return checker


def test_source_current_dir(mockopen, monkeypatch, mocked_execx_checker):
    monkeypatch.setattr(os.path, "isfile", lambda x: True)
    source_alias(["foo", "bar"])
    assert mocked_execx_checker == ["foo", "bar"]


def test_source_path(mockopen, mocked_execx_checker, patch_locate_binary, xession):
    patch_locate_binary(xession.commands_cache)

    source_alias(["foo", "bar"])
    path_foo = os.path.join("bin", "foo")
    path_bar = os.path.join("bin", "bar")
    assert mocked_execx_checker[0].endswith(path_foo)
    assert mocked_execx_checker[1].endswith(path_bar)


@pytest.mark.parametrize(
    "alias",
    [
        "source-bash",
        "source-zsh",
    ],
)
def test_source_foreign_fn_parser(alias, xession):
    aliases = make_default_aliases()
    source_bash = aliases[alias]

    positionals = [act.dest for act in source_bash.parser._get_positional_actions()]
    options = [act.dest for act in source_bash.parser._get_optional_actions()]

    assert positionals == ["files_or_code"]
    assert options == [
        "help",
        "interactive",
        "login",
        "envcmd",
        "aliascmd",
        "extra_args",
        "safe",
        "prevcmd",
        "postcmd",
        "funcscmd",
        "sourcer",
        "use_tmpfile",
        "seterrprevcmd",
        "seterrpostcmd",
        "overwrite_aliases",
        "suppress_skip_message",
        "show",
        "dryrun",
    ]



import os
import subprocess as sp
from pathlib import Path
from unittest.mock import Mock

import pytest

from xonsh.prompt import vc

# Xonsh interaction with version control systems.
VC_BRANCH = {"git": {"master", "main"}, "hg": {"default"}}


@pytest.fixture(params=VC_BRANCH.keys())
def repo(request, tmpdir_factory):
    """Return a dict with vc and a temporary dir
    that is a repository for testing.
    """
    vc = request.param
    temp_dir = Path(tmpdir_factory.mktemp("dir"))
    os.chdir(temp_dir)
    try:
        sp.call([vc, "init"])
    except FileNotFoundError:
        pytest.skip(f"cannot find {vc} executable")
    if vc == "git":
        git_config = temp_dir / ".git/config"
        git_config.write_text(
            """
[user]
name = me
email = my@email.address
[init]
defaultBranch = main
"""
        )

        # git needs at least one commit
        Path("test-file").touch()
        sp.call(["git", "add", "test-file"])
        sp.call(["git", "commit", "-m", "test commit"])
    return {"vc": vc, "dir": temp_dir}


@pytest.fixture
def set_xenv(xession, monkeypatch):
    def _wrapper(path):
        xession.env.update(dict(VC_BRANCH_TIMEOUT=2, PWD=path))
        return xession

    return _wrapper


def test_test_repo(repo):
    test_vc_dir = repo["dir"] / ".{}".format(repo["vc"])
    assert test_vc_dir.is_dir()
    if repo["vc"] == "git":
        test_file = repo["dir"] / "test-file"
        assert test_file.exists()


def test_no_repo(tmpdir, set_xenv):
    set_xenv(tmpdir)
    assert vc.get_hg_branch() is None
    assert vc.get_git_branch() is None


def test_vc_get_branch(repo, set_xenv):
    set_xenv(repo["dir"])
    # get corresponding function from vc module
    get_branch = "get_{}_branch".format(repo["vc"])
    branch = getattr(vc, get_branch)()

    assert branch in VC_BRANCH[repo["vc"]]
    if repo["vc"] == "git":
        git_config = repo["dir"] / ".git/config"
        with git_config.open("a") as f:
            f.write(
                """
[color]
branch = always
interactive = always
[color "branch"]
current = yellow reverse
"""
            )

        branch = getattr(vc, get_branch)()
        assert branch in VC_BRANCH[repo["vc"]]
        assert not branch.startswith("\u001b[")


def test_current_branch_calls_locate_binary_for_empty_cmds_cache(xession, monkeypatch):
    cache = xession.commands_cache
    monkeypatch.setattr(cache, "is_empty", Mock(return_value=True))
    monkeypatch.setattr(cache, "locate_binary", Mock(return_value=""))
    vc.current_branch()
    assert cache.locate_binary.called


def test_current_branch_does_not_call_locate_binary_for_non_empty_cmds_cache(
    xession,
    monkeypatch,
):
    cache = xession.commands_cache
    monkeypatch.setattr(cache, "is_empty", Mock(return_value=False))
    monkeypatch.setattr(cache, "locate_binary", Mock(return_value=""))
    # make lazy locate return nothing to avoid running vc binaries
    monkeypatch.setattr(cache, "lazy_locate_binary", Mock(return_value=""))
    vc.current_branch()
    assert not cache.locate_binary.called


def test_dirty_working_directory(repo, set_xenv):
    get_dwd = "{}_dirty_working_directory".format(repo["vc"])
    set_xenv(repo["dir"])

    # By default, git / hg do not care about untracked files
    Path("second-test-file").touch()
    assert not getattr(vc, get_dwd)()

    sp.call([repo["vc"], "add", "second-test-file"])
    assert getattr(vc, get_dwd)()


@pytest.mark.parametrize("include_untracked", [True, False])
def test_git_dirty_working_directory_includes_untracked(
    xession, fake_process, include_untracked
):
    xession.env["VC_GIT_INCLUDE_UNTRACKED"] = include_untracked

    if include_untracked:
        fake_process.register_subprocess(
            command="git status --porcelain --untracked-files=normal".split(),
            stdout=b"?? untracked-test-file",
        )
    else:
        fake_process.register_subprocess(
            command="git status --porcelain --untracked-files=no".split(),
            stdout=b"",
        )

    assert vc.git_dirty_working_directory() == include_untracked



"""The prompt_toolkit based xonsh shell."""
import os
import re
import sys
from functools import wraps
from types import MethodType

from prompt_toolkit import ANSI
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from prompt_toolkit.enums import EditingMode
from prompt_toolkit.formatted_text import PygmentsTokens, to_formatted_text
from prompt_toolkit.history import ThreadedHistory
from prompt_toolkit.key_binding.bindings.emacs import (
    load_emacs_shift_selection_bindings,
)
from prompt_toolkit.key_binding.bindings.named_commands import get_by_name
from prompt_toolkit.key_binding.key_bindings import merge_key_bindings
from prompt_toolkit.shortcuts import CompleteStyle
from prompt_toolkit.shortcuts import print_formatted_text as ptk_print
from prompt_toolkit.shortcuts.prompt import PromptSession
from prompt_toolkit.styles import Style, merge_styles
from prompt_toolkit.styles.pygments import pygments_token_to_classname

from xonsh.base_shell import BaseShell
from xonsh.built_ins import XSH
from xonsh.events import events
from xonsh.lazyimps import pyghooks, pygments, winutils
from xonsh.platform import HAS_PYGMENTS, ON_POSIX, ON_WINDOWS
from xonsh.ptk_shell.completer import PromptToolkitCompleter
from xonsh.ptk_shell.formatter import PTKPromptFormatter
from xonsh.ptk_shell.history import PromptToolkitHistory, _cust_history_matches
from xonsh.ptk_shell.key_bindings import load_xonsh_bindings
from xonsh.pygments_cache import get_all_styles
from xonsh.shell import transform_command
from xonsh.style_tools import DEFAULT_STYLE_DICT, _TokenType, partial_color_tokenize
from xonsh.tools import carriage_return, print_exception, print_warning

try:
    from prompt_toolkit.clipboard import DummyClipboard
    from prompt_toolkit.clipboard.pyperclip import PyperclipClipboard

    HAVE_SYS_CLIPBOARD = True
except ImportError:
    HAVE_SYS_CLIPBOARD = False

try:
    from prompt_toolkit.cursor_shapes import ModalCursorShapeConfig

    HAVE_CURSOR_SHAPE = True
except ImportError:
    HAVE_CURSOR_SHAPE = False

CAPITAL_PATTERN = re.compile(r"([a-z])([A-Z])")
Token = _TokenType()

events.transmogrify("on_ptk_create", "LoadEvent")
events.doc(
    "on_ptk_create",
    """
on_ptk_create(prompter: PromptSession, history: PromptToolkitHistory, completer: PromptToolkitCompleter, bindings: KeyBindings) ->

Fired after prompt toolkit has been initialized
""",
)


def tokenize_ansi(tokens):
    """Checks a list of (token, str) tuples for ANSI escape sequences and
    extends the token list with the new formatted entries.
    During processing tokens are converted to ``prompt_toolkit.FormattedText``.
    Returns a list of similar (token, str) tuples.
    """
    formatted_tokens = to_formatted_text(tokens)
    ansi_tokens = []
    for style, text in formatted_tokens:
        if "\x1b" in text:
            formatted_ansi = to_formatted_text(ANSI(text))
            ansi_text = ""
            prev_style = ""
            for ansi_style, ansi_text_part in formatted_ansi:
                if prev_style == ansi_style:
                    ansi_text += ansi_text_part
                else:
                    ansi_tokens.append((prev_style or style, ansi_text))
                    prev_style = ansi_style
                    ansi_text = ansi_text_part
            ansi_tokens.append((prev_style or style, ansi_text))
        else:
            ansi_tokens.append((style, text))
    return ansi_tokens


def _pygments_token_to_classname(token):
    """Converts pygments Tokens, token names (strings) to PTK style names."""
    if token and isinstance(token, str):
        # if starts with non capital letter => leave it as it is
        if token[0].islower():
            return token
        # if starts with capital letter => pygments token name
        if token.startswith("Token."):
            token = token[6:]
        # short colors - all caps
        if token == token.upper():
            token = "color." + token
        return "pygments." + token.lower()

    return pygments_token_to_classname(token)


def _style_from_pygments_dict(pygments_dict):
    """Custom implementation of ``style_from_pygments_dict`` that supports PTK specific
    (``Token.PTK``) styles.
    """
    pygments_style = []

    for token, style in pygments_dict.items():
        # if ``Token.PTK`` then add it as "native" PTK style too
        if str(token).startswith("Token.PTK"):
            key = CAPITAL_PATTERN.sub(r"\1-\2", str(token)[10:]).lower()
            pygments_style.append((key, style))
        pygments_style.append((_pygments_token_to_classname(token), style))

    return Style(pygments_style)


def _style_from_pygments_cls(pygments_cls):
    """Custom implementation of ``style_from_pygments_cls`` that supports PTK specific
    (``Token.PTK``) styles.
    """
    return _style_from_pygments_dict(pygments_cls.styles)


def disable_copy_on_deletion():
    dummy_clipboard = DummyClipboard()
    ignored_actions = [
        "kill-line",
        "kill-word",
        "unix-word-rubout",
        "unix-line-discard",
        "backward-kill-word",
    ]

    def handle_binding(name):
        try:
            binding = get_by_name(name)
        except KeyError:
            print_warning(f"Failed to disable clipboard for ptk action {name!r}")
            return

        if getattr(binding, "xonsh_disabled_clipboard", False):
            # binding's clipboard has already been disabled
            return

        binding.xonsh_disabled_clipboard = True
        original_handler = binding.handler

        # this needs to be defined inside a function so that ``binding`` will be the correct one
        @wraps(original_handler)
        def wrapped_handler(event):
            app = event.app
            prev = app.clipboard
            app.clipboard = dummy_clipboard
            try:
                return original_handler(event)
            finally:
                app.clipboard = prev

        binding.handler = wrapped_handler

    for _name in ignored_actions:
        handle_binding(_name)


class PromptToolkitShell(BaseShell):
    """The xonsh shell for prompt_toolkit v2 and later."""

    completion_displays_to_styles = {
        "multi": CompleteStyle.MULTI_COLUMN,
        "single": CompleteStyle.COLUMN,
        "readline": CompleteStyle.READLINE_LIKE,
        "none": None,
    }

    def __init__(self, **kwargs):
        ptk_args = kwargs.pop("ptk_args", {})
        super().__init__(**kwargs)
        if ON_WINDOWS:
            winutils.enable_virtual_terminal_processing()
        self._first_prompt = True
        self.history = ThreadedHistory(PromptToolkitHistory())
        self.push = self._push

        ptk_args.setdefault("history", self.history)
        if not XSH.env.get("XONSH_COPY_ON_DELETE", False):
            disable_copy_on_deletion()
        if HAVE_SYS_CLIPBOARD:
            ptk_args.setdefault("clipboard", PyperclipClipboard())
        self.prompter: PromptSession = PromptSession(**ptk_args)

        self.prompt_formatter = PTKPromptFormatter(self)
        self.pt_completer = PromptToolkitCompleter(self.completer, self.ctx, self)
        ptk_bindings = self.prompter.app.key_bindings
        self.key_bindings = load_xonsh_bindings(ptk_bindings)
        self._overrides_deprecation_warning_shown = False

        # Store original `_history_matches` in case we need to restore it
        self._history_matches_orig = self.prompter.default_buffer._history_matches
        # This assumes that PromptToolkitShell is a singleton
        events.on_ptk_create.fire(
            prompter=self.prompter,
            history=self.history,
            completer=self.pt_completer,
            bindings=self.key_bindings,
        )
        # Goes at the end, since _MergedKeyBindings objects do not have
        # an add() function, which is necessary for on_ptk_create events
        self.key_bindings = merge_key_bindings(
            [self.key_bindings, load_emacs_shift_selection_bindings()]
        )

    def get_lazy_ptk_kwargs(self):
        """These are non-essential attributes for the PTK shell to start.
        Lazy loading these later would save some startup time.
        """
        if not XSH.env.get("COLOR_INPUT"):
            return

        if HAS_PYGMENTS:
            # these imports slowdown a little
            from prompt_toolkit.lexers import PygmentsLexer

            yield "lexer", PygmentsLexer(pyghooks.XonshLexer)

        events.on_timingprobe.fire(name="on_pre_prompt_style")
        yield "style", self.get_prompt_style()
        events.on_timingprobe.fire(name="on_post_prompt_style")

    def get_prompt_style(self):
        env = XSH.env

        style_overrides_env = env.get("PTK_STYLE_OVERRIDES", {}).copy()
        if (
            len(style_overrides_env) > 0
            and not self._overrides_deprecation_warning_shown
        ):
            print_warning(
                "$PTK_STYLE_OVERRIDES is deprecated, use $XONSH_STYLE_OVERRIDES instead!"
            )
            self._overrides_deprecation_warning_shown = True
        style_overrides_env.update(env.get("XONSH_STYLE_OVERRIDES", {}))

        if HAS_PYGMENTS:
            style = _style_from_pygments_cls(pyghooks.xonsh_style_proxy(self.styler))
            if len(self.styler.non_pygments_rules) > 0:
                try:
                    style = merge_styles(
                        [
                            style,
                            _style_from_pygments_dict(self.styler.non_pygments_rules),
                        ]
                    )
                except (AttributeError, TypeError, ValueError) as style_exception:
                    print_warning(
                        f"Error applying style override!\n{style_exception}\n"
                    )

        else:
            style = _style_from_pygments_dict(DEFAULT_STYLE_DICT)

        if len(style_overrides_env) > 0:
            try:
                style = merge_styles(
                    [style, _style_from_pygments_dict(style_overrides_env)]
                )
            except (AttributeError, TypeError, ValueError) as style_exception:
                print_warning(f"Error applying style override!\n{style_exception}\n")
        return style

    def singleline(
        self, auto_suggest=None, enable_history_search=True, multiline=True, **kwargs
    ):
        """Reads a single line of input from the shell. The store_in_history
        kwarg flags whether the input should be stored in PTK's in-memory
        history.
        """
        events.on_pre_prompt_format.fire()
        env = XSH.env
        mouse_support = env.get("MOUSE_SUPPORT")
        auto_suggest = auto_suggest if env.get("AUTO_SUGGEST") else None
        refresh_interval = env.get("PROMPT_REFRESH_INTERVAL")
        refresh_interval = refresh_interval if refresh_interval > 0 else None
        complete_in_thread = env.get("COMPLETION_IN_THREAD")
        completions_display = env.get("COMPLETIONS_DISPLAY")
        complete_style = self.completion_displays_to_styles[completions_display]

        complete_while_typing = env.get("UPDATE_COMPLETIONS_ON_KEYPRESS")
        if complete_while_typing:
            # PTK requires history search to be none when completing while typing
            enable_history_search = False
        if HAS_PYGMENTS:
            self.styler.style_name = env.get("XONSH_COLOR_STYLE")
        completer = None if completions_display == "none" else self.pt_completer

        events.on_timingprobe.fire(name="on_pre_prompt_tokenize")

        # clear prompt level cache
        env["PROMPT_FIELDS"].reset()

        get_bottom_toolbar_tokens = self.bottom_toolbar_tokens
        if env.get("UPDATE_PROMPT_ON_KEYPRESS"):
            get_prompt_tokens = self.prompt_tokens
            get_rprompt_tokens = self.rprompt_tokens
        else:
            get_prompt_tokens = self.prompt_tokens()
            get_rprompt_tokens = self.rprompt_tokens()
            if get_bottom_toolbar_tokens:
                get_bottom_toolbar_tokens = get_bottom_toolbar_tokens()
        events.on_timingprobe.fire(name="on_post_prompt_tokenize")

        if env.get("VI_MODE"):
            editing_mode = EditingMode.VI
        else:
            editing_mode = EditingMode.EMACS

        if env.get("XONSH_HISTORY_MATCH_ANYWHERE"):
            self.prompter.default_buffer._history_matches = MethodType(
                _cust_history_matches, self.prompter.default_buffer
            )
        elif (
            self.prompter.default_buffer._history_matches
            is not self._history_matches_orig
        ):
            self.prompter.default_buffer._history_matches = self._history_matches_orig

        menu_rows = env.get("COMPLETIONS_MENU_ROWS", None)
        if menu_rows:
            # https://github.com/xonsh/xonsh/pull/4477#pullrequestreview-767982976
            menu_rows += 1

        prompt_args = {
            "mouse_support": mouse_support,
            "auto_suggest": auto_suggest,
            "message": get_prompt_tokens,
            "rprompt": get_rprompt_tokens,
            "bottom_toolbar": get_bottom_toolbar_tokens,
            "completer": completer,
            "multiline": multiline,
            "editing_mode": editing_mode,
            "prompt_continuation": self.continuation_tokens,
            "enable_history_search": enable_history_search,
            "reserve_space_for_menu": menu_rows,
            "key_bindings": self.key_bindings,
            "complete_style": complete_style,
            "complete_while_typing": complete_while_typing,
            "include_default_pygments_style": False,
            "refresh_interval": refresh_interval,
            "complete_in_thread": complete_in_thread,
        }
        if env["ENABLE_ASYNC_PROMPT"]:
            # once the prompt is done, update it in background as each future is completed
            prompt_args["pre_run"] = self.prompt_formatter.start_update
        else:
            for attr, val in self.get_lazy_ptk_kwargs():
                prompt_args[attr] = val

        if editing_mode == EditingMode.VI and HAVE_CURSOR_SHAPE:
            prompt_args["cursor"] = ModalCursorShapeConfig()
        events.on_pre_prompt.fire()
        line = self.prompter.prompt(**prompt_args)
        events.on_post_prompt.fire()
        return line

    def _push(self, line):
        """Pushes a line onto the buffer and compiles the code in a way that
        enables multiline input.
        """
        code = None
        self.buffer.append(line)
        if self.need_more_lines:
            return None, code
        src = "".join(self.buffer)
        src = transform_command(src)
        try:
            code = self.execer.compile(
                src, mode="single", glbs=self.ctx, locs=None, compile_empty_tree=False
            )
            self.reset_buffer()
        except Exception:  # pylint: disable=broad-except
            self.reset_buffer()
            print_exception()
            return src, None
        return src, code

    def cmdloop(self, intro=None):
        """Enters a loop that reads and execute input from user."""
        if intro:
            print(intro)
        auto_suggest = AutoSuggestFromHistory()
        while not XSH.exit:
            try:
                line = self.singleline(auto_suggest=auto_suggest)
                if not line:
                    self.emptyline()
                else:
                    raw_line = line
                    line = self.precmd(line)
                    self.default(line, raw_line)
            except (KeyboardInterrupt, SystemExit):
                self.reset_buffer()
            except EOFError:
                if XSH.env.get("IGNOREEOF"):
                    print('Use "exit" to leave the shell.', file=sys.stderr)
                else:
                    break

    def _get_prompt_tokens(self, env_name: str, prompt_name: str, **kwargs):
        env = XSH.env  # type:ignore
        p = env.get(env_name)

        if not p and "default" in kwargs:
            return kwargs.pop("default")

        try:
            p = self.prompt_formatter(
                template=p, threaded=env["ENABLE_ASYNC_PROMPT"], prompt_name=prompt_name
            )
        except Exception:  # pylint: disable=broad-except
            print_exception()

        toks = partial_color_tokenize(p)

        return tokenize_ansi(PygmentsTokens(toks))

    def prompt_tokens(self):
        """Returns a list of (token, str) tuples for the current prompt."""
        if self._first_prompt:
            carriage_return()
            self._first_prompt = False

        tokens = self._get_prompt_tokens("PROMPT", "message")
        self.settitle()
        return tokens

    def rprompt_tokens(self):
        """Returns a list of (token, str) tuples for the current right
        prompt.
        """
        return self._get_prompt_tokens("RIGHT_PROMPT", "rprompt", default=[])

    def _bottom_toolbar_tokens(self):
        """Returns a list of (token, str) tuples for the current bottom
        toolbar.
        """
        return self._get_prompt_tokens("BOTTOM_TOOLBAR", "bottom_toolbar", default=None)

    @property
    def bottom_toolbar_tokens(self):
        """Returns self._bottom_toolbar_tokens if it would yield a result"""
        if XSH.env.get("BOTTOM_TOOLBAR"):
            return self._bottom_toolbar_tokens

    def continuation_tokens(self, width, line_number, is_soft_wrap=False):
        """Displays dots in multiline prompt"""
        if is_soft_wrap:
            return ""
        width -= 1
        dots = XSH.env.get("MULTILINE_PROMPT")
        dots = dots() if callable(dots) else dots
        if not dots:
            return ""
        basetoks = self.format_color(dots)
        baselen = sum(len(t[1]) for t in basetoks)
        if baselen == 0:
            return [(Token, " " * (width + 1))]
        toks = basetoks * (width // baselen)
        n = width % baselen
        count = 0
        for tok in basetoks:
            slen = len(tok[1])
            newcount = slen + count
            if slen == 0:
                continue
            elif newcount <= n:
                toks.append(tok)
            else:
                toks.append((tok[0], tok[1][: n - count]))
            count = newcount
            if n <= count:
                break
        toks.append((Token, " "))  # final space
        return PygmentsTokens(toks)

    def format_color(self, string, hide=False, force_string=False, **kwargs):
        """Formats a color string using Pygments. This, therefore, returns
        a list of (Token, str) tuples. If force_string is set to true, though,
        this will return a color formatted string.
        """
        tokens = partial_color_tokenize(string)
        if force_string and HAS_PYGMENTS:
            env = XSH.env
            style_overrides_env = env.get("XONSH_STYLE_OVERRIDES", {})
            self.styler.style_name = env.get("XONSH_COLOR_STYLE")
            self.styler.override(style_overrides_env)
            proxy_style = pyghooks.xonsh_style_proxy(self.styler)
            formatter = pyghooks.XonshTerminal256Formatter(style=proxy_style)
            s = pygments.format(tokens, formatter)
            return s
        elif force_string:
            print("To force colorization of string, install Pygments")
            return tokens
        else:
            return tokens

    def print_color(self, string, end="\n", **kwargs):
        """Prints a color string using prompt-toolkit color management."""
        if isinstance(string, str):
            tokens = partial_color_tokenize(string)
        else:
            # assume this is a list of (Token, str) tuples and just print
            tokens = string
        tokens = PygmentsTokens(tokens)
        env = XSH.env
        style_overrides_env = env.get("XONSH_STYLE_OVERRIDES", {})
        if HAS_PYGMENTS:
            self.styler.style_name = env.get("XONSH_COLOR_STYLE")
            self.styler.override(style_overrides_env)
            proxy_style = _style_from_pygments_cls(
                pyghooks.xonsh_style_proxy(self.styler)
            )
        else:
            proxy_style = merge_styles(
                [
                    _style_from_pygments_dict(DEFAULT_STYLE_DICT),
                    _style_from_pygments_dict(style_overrides_env),
                ]
            )
        ptk_print(
            tokens,
            style=proxy_style,
            end=end,
            include_default_pygments_style=False,
            **kwargs,
        )

    def color_style_names(self):
        """Returns an iterable of all available style names."""
        if not HAS_PYGMENTS:
            return ["For other xonsh styles, please install pygments"]
        return get_all_styles()

    def color_style(self):
        """Returns the current color map."""
        if not HAS_PYGMENTS:
            return DEFAULT_STYLE_DICT
        env = XSH.env
        self.styler.style_name = env.get("XONSH_COLOR_STYLE")
        return self.styler.styles

    def restore_tty_sanity(self):
        """An interface for resetting the TTY stdin mode. This is highly
        dependent on the shell backend. Also it is mostly optional since
        it only affects ^Z backgrounding behaviour.
        """
        # PTK does not seem to need any specialization here. However,
        # if it does for some reason in the future...
        # The following writes an ANSI escape sequence that sends the cursor
        # to the end of the line. This has the effect of restoring ECHO mode.
        # See http://unix.stackexchange.com/a/108014/129048 for more details.
        # This line can also be replaced by os.system("stty sane"), as per
        # http://stackoverflow.com/questions/19777129/interactive-python-interpreter-run-in-background#comment29421919_19778355
        # However, it is important to note that not termios-based solution
        # seems to work. My guess is that this is because termios restoration
        # needs to be performed by the subprocess itself. This fix is important
        # when subprocesses don't properly restore the terminal attributes,
        # like Python in interactive mode. Also note that the sequences "\033M"
        # and "\033E" seem to work too, but these are technically VT100 codes.
        # I used the more primitive ANSI sequence to maximize compatibility.
        # -scopatz 2017-01-28
        #   if not ON_POSIX:
        #       return
        #   sys.stdout.write('\033[9999999C\n')
        if not ON_POSIX:
            return
        stty, _ = XSH.commands_cache.lazyget("stty", (None, None))
        if stty is None:
            return
        os.system(stty + " sane")



"""Simple built-in debugger. Runs pdb on reception of SIGUSR1 signal."""
import signal

from xonsh.built_ins import XonshSession


def handle_sigusr1(sig, frame):
    print("\nSIGUSR1 signal received. Starting interactive debugger...", flush=True)
    import pdb

    pdb.Pdb().set_trace(frame)


def _load_xontrib_(xsh: XonshSession, **_):
    signal.signal(signal.SIGUSR1, handle_sigusr1)



"""Vox tests"""
import io
import os
import pathlib
import stat
import subprocess as sp
import sys
import types
from typing import TYPE_CHECKING

import pytest
from py.path import local

from xonsh.platform import ON_WINDOWS
from xonsh.pytest.tools import skip_if_on_conda, skip_if_on_msys
from xontrib.voxapi import Vox, _get_vox_default_interpreter

if TYPE_CHECKING:
    from pytest_subprocess import FakeProcess

    from xontrib.vox import VoxHandler


@pytest.fixture
def venv_home(tmpdir, xession):
    """Path where VENVs are created"""
    home = tmpdir / "venvs"
    home.ensure_dir()
    # Set up an isolated venv home
    xession.env["VIRTUALENV_HOME"] = str(home)
    return home


@pytest.fixture
def venv_proc(fake_process: "FakeProcess", venv_home):
    def version_handle(process):
        ver = str(sys.version).split()[0]
        process.stdout.write(f"Python {ver}")

    def venv_handle(process):
        env_path = local(process.args[3])
        (env_path / "lib").ensure_dir()
        bin_path = env_path / ("Scripts" if ON_WINDOWS else "bin")

        (bin_path / "python").write("", ensure=True)
        (bin_path / "python.exe").write("", ensure=True)
        for file in bin_path.listdir():
            st = os.stat(str(file))
            os.chmod(str(file), st.st_mode | stat.S_IEXEC)

        for pip_name in ["pip", "pip.exe"]:
            fake_process.register(
                [str(bin_path / pip_name), "freeze", fake_process.any()], stdout=""
            )

            # will be used by `vox runin`
            fake_process.register(
                [pip_name, "--version"],
                stdout=f"pip 22.0.4 from {env_path}/lib/python3.10/site-packages/pip (python 3.10)",
            )
        fake_process.keep_last_process(True)
        return env_path

    def get_interpreters():
        interpreter = _get_vox_default_interpreter()
        yield interpreter
        if sys.executable != interpreter:
            yield sys.executable

    for cmd in get_interpreters():
        fake_process.register([cmd, "--version"], callback=version_handle)
        venv = (cmd, "-m", "venv")
        fake_process.register([*venv, fake_process.any(min=1)], callback=venv_handle)
    fake_process.keep_last_process(True)
    return fake_process


@pytest.fixture
def vox(xession, load_xontrib, venv_proc) -> "VoxHandler":
    """vox Alias function"""

    # Set up enough environment for xonsh to function
    xession.env["PWD"] = os.getcwd()
    xession.env["DIRSTACK_SIZE"] = 10
    xession.env["PATH"] = []
    xession.env["XONSH_SHOW_TRACEBACK"] = True

    load_xontrib("vox")
    vox = xession.aliases["vox"]
    return vox


@pytest.fixture
def record_events(xession):
    class Listener:
        def __init__(self):
            self.last = None

        def listener(self, name):
            def _wrapper(**kwargs):
                self.last = (name,) + tuple(kwargs.values())

            return _wrapper

        def __call__(self, *events: str):
            for name in events:
                event = getattr(xession.builtins.events, name)
                event(self.listener(name))

    yield Listener()


def test_vox_flow(xession, vox, record_events, venv_home):
    """
    Creates a virtual environment, gets it, enumerates it, and then deletes it.
    """

    record_events(
        "vox_on_create", "vox_on_delete", "vox_on_activate", "vox_on_deactivate"
    )

    vox(["create", "spam"])
    assert stat.S_ISDIR(venv_home.join("spam").stat().mode)
    assert record_events.last == ("vox_on_create", "spam")

    ve = vox.vox["spam"]
    assert ve.env == str(venv_home.join("spam"))
    assert os.path.isdir(ve.bin)

    assert "spam" in vox.vox
    assert "spam" in list(vox.vox)

    # activate
    vox(["activate", "spam"])
    assert xession.env["VIRTUAL_ENV"] == vox.vox["spam"].env
    assert record_events.last == ("vox_on_activate", "spam", str(ve.env))

    out = io.StringIO()
    # info
    vox(["info"], stdout=out)
    assert "spam" in out.getvalue()
    out.seek(0)

    # list
    vox(["list"], stdout=out)
    print(out.getvalue())
    assert "spam" in out.getvalue()
    out.seek(0)

    # wipe
    vox(["wipe"], stdout=out)
    print(out.getvalue())
    assert "Nothing to remove" in out.getvalue()
    out.seek(0)

    # deactivate
    vox(["deactivate"])
    assert "VIRTUAL_ENV" not in xession.env
    assert record_events.last == ("vox_on_deactivate", "spam", str(ve.env))

    # runin
    vox(["runin", "spam", "pip", "--version"], stdout=out)
    print(out.getvalue())
    assert "spam" in out.getvalue()
    out.seek(0)

    # removal
    vox(["rm", "spam", "--force"])
    assert not venv_home.join("spam").check()
    assert record_events.last == ("vox_on_delete", "spam")


def test_activate_non_vox_venv(xession, vox, record_events, venv_proc, venv_home):
    """
    Create a virtual environment using Python's built-in venv module
    (not in VIRTUALENV_HOME) and verify that vox can activate it correctly.
    """
    xession.env["PATH"] = []

    record_events("vox_on_activate", "vox_on_deactivate")

    with venv_home.as_cwd():
        venv_dirname = "venv"
        sp.run([sys.executable, "-m", "venv", venv_dirname])
        vox(["activate", venv_dirname])
        vxv = vox.vox[venv_dirname]

    env = xession.env
    assert os.path.isabs(vxv.bin)
    assert env["PATH"][0] == vxv.bin
    assert os.path.isabs(vxv.env)
    assert env["VIRTUAL_ENV"] == vxv.env
    assert record_events.last == (
        "vox_on_activate",
        venv_dirname,
        str(pathlib.Path(str(venv_home)) / venv_dirname),
    )

    vox(["deactivate"])
    assert not env["PATH"]
    assert "VIRTUAL_ENV" not in env
    assert record_events.last == (
        "vox_on_deactivate",
        venv_dirname,
        str(pathlib.Path(str(venv_home)) / venv_dirname),
    )


@skip_if_on_msys
@skip_if_on_conda
def test_path(xession, vox, a_venv):
    """
    Test to make sure Vox properly activates and deactivates by examining $PATH
    """
    oldpath = list(xession.env["PATH"])
    vox(["activate", a_venv.basename])

    assert oldpath != xession.env["PATH"]

    vox.deactivate()

    assert oldpath == xession.env["PATH"]


def test_crud_subdir(xession, venv_home, venv_proc):
    """
    Creates a virtual environment, gets it, enumerates it, and then deletes it.
    """

    vox = Vox(force_removals=True)
    vox.create("spam/eggs")
    assert stat.S_ISDIR(venv_home.join("spam", "eggs").stat().mode)

    ve = vox["spam/eggs"]
    assert ve.env == str(venv_home.join("spam", "eggs"))
    assert os.path.isdir(ve.bin)

    assert "spam/eggs" in vox
    assert "spam" not in vox

    # assert 'spam/eggs' in list(vox)  # This is NOT true on Windows
    assert "spam" not in list(vox)

    del vox["spam/eggs"]

    assert not venv_home.join("spam", "eggs").check()


def test_crud_path(xession, tmpdir, venv_proc):
    """
    Creates a virtual environment, gets it, enumerates it, and then deletes it.
    """
    tmp = str(tmpdir)

    vox = Vox(force_removals=True)
    vox.create(tmp)
    assert stat.S_ISDIR(tmpdir.join("lib").stat().mode)

    ve = vox[tmp]
    assert ve.env == str(tmp)
    assert os.path.isdir(ve.bin)

    del vox[tmp]

    assert not tmpdir.check()


@skip_if_on_msys
@skip_if_on_conda
def test_reserved_names(xession, tmpdir):
    """
    Tests that reserved words are disallowed.
    """
    xession.env["VIRTUALENV_HOME"] = str(tmpdir)

    vox = Vox()
    with pytest.raises(ValueError):
        if ON_WINDOWS:
            vox.create("Scripts")
        else:
            vox.create("bin")

    with pytest.raises(ValueError):
        if ON_WINDOWS:
            vox.create("spameggs/Scripts")
        else:
            vox.create("spameggs/bin")


@pytest.mark.parametrize("registered", [False, True])
def test_autovox(xession, vox, a_venv, load_xontrib, registered):
    """
    Tests that autovox works
    """
    from xonsh.dirstack import popd, pushd

    # Makes sure that event handlers are registered
    load_xontrib("autovox")

    env_name = a_venv.basename
    env_path = str(a_venv)

    # init properly
    assert vox.parser

    def policy(path, **_):
        if str(path) == env_path:
            return env_name

    if registered:
        xession.builtins.events.autovox_policy(policy)

    pushd([env_path])
    value = env_name if registered else None
    assert vox.vox.active() == value
    popd([])


@pytest.fixture
def create_venv(venv_proc):
    vox = Vox(force_removals=True)

    def wrapped(name):
        vox.create(name)
        return local(vox[name].env)

    return wrapped


@pytest.fixture
def venvs(venv_home, create_venv):
    """Create virtualenv with names venv0, venv1"""
    from xonsh.dirstack import popd, pushd

    pushd([str(venv_home)])
    yield [create_venv(f"venv{idx}") for idx in range(2)]
    popd([])


@pytest.fixture
def a_venv(create_venv):
    return create_venv("venv0")


@pytest.fixture
def patched_cmd_cache(xession, vox, monkeypatch):
    cc = xession.commands_cache

    def no_change(self, *_):
        return False, False, False

    monkeypatch.setattr(cc, "_check_changes", types.MethodType(no_change, cc))
    monkeypatch.setattr(cc, "_update_cmds_cache", types.MethodType(no_change, cc))
    bins = {path: (path, False) for path in _PY_BINS}
    cc._cmds_cache = bins
    yield cc


_VENV_NAMES = {"venv1", "venv1/", "venv0/", "venv0"}
if ON_WINDOWS:
    _VENV_NAMES = {"venv1\\", "venv0\\", "venv0", "venv1"}

_HELP_OPTS = {
    "-h",
    "--help",
}
_PY_BINS = {"/bin/python2", "/bin/python3"}

_VOX_NEW_OPTS = {
    "--ssp",
    "--system-site-packages",
    "--without-pip",
}.union(_HELP_OPTS)

if ON_WINDOWS:
    _VOX_NEW_OPTS.add("--symlinks")
else:
    _VOX_NEW_OPTS.add("--copies")

_VOX_RM_OPTS = {"-f", "--force"}.union(_HELP_OPTS)


class TestVoxCompletions:
    @pytest.fixture
    def check(self, check_completer, xession, vox):
        def wrapped(cmd, positionals, options=None):
            for k in list(xession.completers):
                if k != "alias":
                    xession.completers.pop(k)
            assert check_completer(cmd) == positionals
            xession.env["ALIAS_COMPLETIONS_OPTIONS_BY_DEFAULT"] = True
            if options:
                assert check_completer(cmd) == positionals.union(options)

        return wrapped

    @pytest.mark.parametrize(
        "args, positionals, opts",
        [
            (
                "vox",
                {
                    "delete",
                    "new",
                    "remove",
                    "del",
                    "workon",
                    "list",
                    "exit",
                    "info",
                    "ls",
                    "rm",
                    "deactivate",
                    "activate",
                    "enter",
                    "create",
                    "project-get",
                    "project-set",
                    "runin",
                    "runin-all",
                    "toggle-ssp",
                    "wipe",
                    "upgrade",
                },
                _HELP_OPTS,
            ),
            (
                "vox create",
                set(),
                _VOX_NEW_OPTS.union(
                    {
                        "-a",
                        "--activate",
                        "--wp",
                        "--without-pip",
                        "-p",
                        "--interpreter",
                        "-i",
                        "--install",
                        "-l",
                        "--link",
                        "--link-project",
                        "-r",
                        "--requirements",
                        "-t",
                        "--temp",
                        "--prompt",
                    }
                ),
            ),
            ("vox activate", _VENV_NAMES, _HELP_OPTS.union({"-n", "--no-cd"})),
            ("vox rm", _VENV_NAMES, _VOX_RM_OPTS),
            ("vox rm venv1", _VENV_NAMES, _VOX_RM_OPTS),  # pos nargs: one or more
            ("vox rm venv1 venv2", _VENV_NAMES, _VOX_RM_OPTS),  # pos nargs: two or more
        ],
    )
    def test_vox_commands(self, args, positionals, opts, check, venvs):
        check(args, positionals, opts)

    @pytest.mark.parametrize(
        "args",
        [
            "vox new --activate --interpreter",  # option after option
            "vox new --interpreter",  # "option: first
            "vox new --activate env1 --interpreter",  # option after pos
            "vox new env1 --interpreter",  # "option: at end"
            "vox new env1 --interpreter=",  # "option: at end with
        ],
    )
    def test_interpreter(self, check, args, patched_cmd_cache):
        check(args, _PY_BINS)



"""
This adds `xog` - a simple command to establish and print temporary traceback log file.
"""

import os
import pathlib
import tempfile

from xonsh.built_ins import XSH, XonshSession


def _get_log_file_name():
    return pathlib.Path(f"{tempfile.gettempdir()}/xonsh-{os.getpid()}.log")


def _clear_log(logfile, stderr):
    try:
        os.remove(logfile)
    except OSError as e:
        print(f"xog: {e}", file=stderr)
        return False
    return True


def _print_log(logfile, stdout, stderr):
    try:
        with open(logfile) as log:
            for line in log:
                print(line, end="", file=stdout)
    except OSError as e:
        print(f"xog: {e}", file=stderr)
        return False
    return True


def _print_help(stdout):
    print(
        """Usage: xog [OPTIONS]
Prints contents of the shell's traceback log file.

Options:
  -c, --clear\t\tclear the log file contents
  -?, --help\t\tprint this help text
"""
    )


def _xog(args, stdout=None, stderr=None):
    if "-?" in args or "--help" in args:
        _print_help(stdout)
        return 0

    logfile = XSH.env.get("XONSH_TRACEBACK_LOGFILE", "")
    if not (logfile and os.path.isfile(logfile)):
        print("Traceback log file doesn't exist.", file=stderr)
        return -1

    if "-c" in args or "--clear" in args:
        rc = _clear_log(logfile, stderr=stderr)
        return 0 if rc else -1

    rc = _print_log(logfile, stdout, stderr)
    return 0 if rc else -1


def _load_xontrib_(xsh: XonshSession, **_):
    xsh.env["XONSH_TRACEBACK_LOGFILE"] = _get_log_file_name()
    xsh.aliases["xog"] = _xog



from xonsh.prompt.cwd import _replace_home_cwd


def test_cwd_escapes_curly_brackets_with_more_curly_brackets(xession, tmpdir):
    xession.env["HOME"] = str(tmpdir)
    xession.env["PWD"] = "{foo}"
    assert _replace_home_cwd() == "{{foo}}"

    xession.env["PWD"] = "{{foo}}"
    assert _replace_home_cwd() == "{{{{foo}}}}"

    xession.env["PWD"] = "{"
    assert _replace_home_cwd() == "{{"

    xession.env["PWD"] = "}}"
    assert _replace_home_cwd() == "}}}}"



import pytest


@pytest.mark.parametrize(
    "args, prefix, exp",
    [
        (
            "xonfig",
            "-",
            {"-h", "--help"},
        ),
        (
            "xonfig colors",
            "b",
            {"blue", "brown"},
        ),
    ],
)
def test_xonfig(args, prefix, exp, xsh_with_aliases, monkeypatch, check_completer):
    from xonsh import xonfig

    monkeypatch.setattr(xonfig, "color_style_names", lambda: ["blue", "brown", "other"])
    assert check_completer(args, prefix=prefix) == exp


@pytest.mark.parametrize(
    "args, prefix, exp, exp_part",
    [
        (
            "xontrib",
            "l",
            {"list", "load"},
            None,
        ),
        (
            "xontrib load",
            "",
            None,
            {
                # the list may vary wrt the env. so testing only part of the coreutils.
                "abbrevs",
                "pdb",
                "bashisms",
                "coreutils",
            },
        ),
    ],
)
def test_xontrib(args, prefix, exp, exp_part, xsh_with_aliases, check_completer):
    result = check_completer(args, prefix=prefix)
    if exp:
        assert result == exp
    if exp_part:
        assert result.issuperset(exp_part), f"{result} doesn't contain {exp_part} "


def test_module_matcher(tmp_path, xession):
    from xonsh.completers import commands

    for idx, ext in enumerate(commands.ModuleFinder.extensions):
        (tmp_path / f"a{idx}{ext}").write_text("def xonsh_complete(): pass")

    matcher = commands.ModuleFinder("xompletions", str(tmp_path))
    assert matcher.get_module("pip").xonsh_complete
    assert matcher.get_module("a0").xonsh_complete
    # todo: fix *.xsh import
    #  the import-hook returns None for some reason
    #  -- xonsh/imphooks.py:247
    # assert matcher.get_module("a1").xonsh_complete
